from composio.client.types import Tool
from composio.core.models.connected_accounts import auth_scheme
from composio.core.models.custom_tools import ExecuteRequestFn
from composio.core.models.tools import (
    Modifiers,
    ToolExecuteParams,
    ToolExecutionResponse,
)
from composio.core.provider.base import TTool, TToolCollection

__all__ = [
    "Tool",
    "TTool",
    "TToolCollection",
    "ToolExecuteParams",
    "ToolExecutionResponse",
    "ExecuteRequestFn",
    "Modifiers",
    "auth_scheme",
]



"""
Google AI Python Gemini demo.
"""

import dotenv

from composio import Composio
from composio_google import GoogleProvider
from vertexai.generative_models import GenerativeModel


# Load environment variables from .env
dotenv.load_dotenv()

# Initialize tools
composio = Composio(provider=GoogleProvider())

# Get GitHub tools that are pre-configured
tool = composio.tools.get(user_id="default", toolkits=["GITHUB"])

# Initialize the Gemini model
model = GenerativeModel("gemini-1.5-pro", tools=[tool])

# Start a chat session
chat = model.start_chat()


def main():
    # Define task
    task = "Star a repo composiohq/composio on GitHub"

    # Send a message to the model
    response = chat.send_message(task)

    print("Model response:")
    print(response)

    result = composio.provider.handle_response(user_id="default", response=response)
    print("Function call result:")
    print(result)


if __name__ == "__main__":
    main()



"""
Anthropic claude demo.7
"""

import anthropic

from composio import Composio
from composio_anthropic import AnthropicProvider


# Initialize tools.
anthropic_client = anthropic.Anthropic()
composio = Composio(provider=AnthropicProvider())

# Get GitHub tools that are pre-configured
tools = composio.tools.get(user_id="default", toolkits=["GITHUB"])

# Get response from the LLM
response = anthropic_client.messages.create(
    model="claude-3-opus-20240229",
    max_tokens=1024,
    tools=tools,
    messages=[
        {"role": "user", "content": "Star me composiohq/composio repo in github."},
    ],
)

# Execute the function calls.
result = composio.provider.handle_tool_calls(user_id="default", response=response)
print(result)



"""
Base resource class for representing resources in the composio client.
"""

import contextvars
import functools
import os
import time
import traceback
import typing as t

from composio.__version__ import __version__
from composio.client import HttpClient
from composio.utils.logging import WithLogger

from ._telemetry import Event, create_event, push_event

PayloadT = t.TypeVar("PayloadT", bound=dict)

allow_tracking = contextvars.ContextVar[bool]("allow_tracking", default=True)
_environment = os.getenv("ENVIRONMENT", "development")


def trace_method(method: t.Callable, name: str, **attributes: t.Any) -> t.Callable:
    """Wrap a method to log the call."""

    @functools.wraps(method)
    def trace_wrapper(self, *args, **kwargs):
        if not allow_tracking.get():
            return method(self, *args, **kwargs)

        event: t.Optional[Event] = None
        start_time = time.time()
        event = create_event(
            type="metric",
            functionName=name,
            timestamp=time.time(),
            props=attributes,
            source={
                "environment": _environment,  # type: ignore
                "language": "python",
                "service": "sdk",
                "version": __version__,
            },
            metadata={
                "provider": self._client.provider,
            },
        )
        try:
            return method(self, *args, **kwargs)
        except Exception as e:
            _, payload = event
            payload["error"] = {
                "name": e.__class__.__name__,
                "message": str(e),
                "stack": traceback.format_exc(),
            }
            event = ("error", payload)
            raise e
        finally:
            if event is not None:
                event[1]["durationMs"] = (time.time() - start_time) * 1000
                push_event(event=event)

    trace_wrapper.__name__ = method.__name__
    return trace_wrapper


class ResourceMeta(type):
    """Meta class for resource classes."""

    def __init__(cls, name, bases, attrs):
        for attr in attrs:
            if "__" in attr or not callable(getattr(cls, attr)):
                continue
            setattr(cls, attr, trace_method(getattr(cls, attr), f"{name}.{attr}"))


class Resource(WithLogger, metaclass=ResourceMeta):
    """Base resource class for composio client."""

    def sanitize_payload(self, payload: PayloadT) -> PayloadT:
        return {k: v for k, v in payload.items()}  # type: ignore

    def __init__(self, client: HttpClient):
        super().__init__()
        self._client = client



import requests
from pydantic import BaseModel, Field

from composio import Composio
from composio.types import ExecuteRequestFn

composio = Composio()


class AddTwoNumbersInput(BaseModel):
    a: int = Field(
        ...,
        description="The first number to add",
    )
    b: int = Field(
        ...,
        description="The second number to add",
    )


@composio.tools.custom_tool
def add_two_numbers(request: AddTwoNumbersInput) -> int:
    """Add two numbers."""
    return request.a + request.b


class GetIssueInfoInput(BaseModel):
    issue_number: int = Field(
        ...,
        description="The number of the issue to get information about",
    )


@composio.tools.custom_tool(toolkit="github")
def get_issue_info(
    request: GetIssueInfoInput,
    execute_request: ExecuteRequestFn,
    auth_credentials: dict,
) -> dict:
    """Get information about a GitHub issue."""
    response = execute_request(
        endpoint=f"/repos/composiohq/composio/issues/{request.issue_number}",
        method="GET",
        parameters=[
            {
                "name": "Accept",
                "value": "application/vnd.github.v3+json",
                "type": "header",
            },
            {
                "name": "Authorization",
                "value": f"Bearer {auth_credentials['access_token']}",
                "type": "header",
            },
        ],
    )
    return {"data": response.data}


@composio.tools.custom_tool(toolkit="github")
def get_issue_info_direct(
    request: GetIssueInfoInput,
    execute_request: ExecuteRequestFn,
    auth_credentials: dict,
) -> dict:
    """Get information about a GitHub issue."""
    response = requests.get(
        f"https://api.github.com/repos/composiohq/composio/issues/{request.issue_number}",
        headers={
            "Accept": "application/vnd.github.v3+json",
            "Authorization": f"Bearer {auth_credentials['access_token']}",
        },
    )
    return {"data": response.json()}


response = composio.tools.execute(
    user_id="default",
    slug=get_issue_info.slug,
    arguments={"issue_number": 1},
)

print(response)



import typing as t
from pathlib import Path

_default = "application/octet-stream"
_types = {
    ".3dm": "x-world/x-3dmf",
    ".3dmf": "x-world/x-3dmf",
    ".7z": "application/x-7z-compressed",
    ".a": "application/octet-stream",
    ".aab": "application/x-authorware-bin",
    ".aam": "application/x-authorware-map",
    ".aas": "application/x-authorware-seg",
    ".abc": "text/vnd.abc",
    ".acgi": "text/html",
    ".afl": "video/animaflex",
    ".ai": "application/postscript",
    ".aif": "audio/x-aiff",
    ".aifc": "audio/x-aiff",
    ".aiff": "audio/x-aiff",
    ".aim": "application/x-aim",
    ".aip": "text/x-audiosoft-intra",
    ".ani": "application/x-navi-animation",
    ".aos": "application/x-nokia-9000-communicator-add-on-software",
    ".aps": "application/mime",
    ".arc": "application/octet-stream",
    ".arj": "application/octet-stream",
    ".art": "image/x-jg",
    ".asf": "video/x-ms-asf",
    ".asm": "text/x-asm",
    ".asp": "text/asp",
    ".asx": "video/x-ms-asf-plugin",
    ".au": "audio/basic",
    ".avi": "video/x-msvideo",
    ".avs": "video/avs-video",
    ".bcpio": "application/x-bcpio",
    ".bin": "application/octet-stream",
    ".bm": "image/bmp",
    ".bmp": "image/x-ms-bmp",
    ".boo": "application/book",
    ".book": "application/book",
    ".boz": "application/x-bzip2",
    ".bsh": "application/x-bsh",
    ".bz": "application/x-bzip",
    ".bz2": "application/x-bzip2",
    ".c": "text/plain",
    ".c+": "text/plain",
    ".cat": "application/vnd.ms-pki.seccat",
    ".cc": "text/x-c",
    ".ccad": "application/clariscad",
    ".cco": "application/x-cocoa",
    ".cdf": "application/x-netcdf",
    ".cer": "application/x-x509-ca-cert",
    ".cha": "application/x-chat",
    ".chat": "application/x-chat",
    ".class": "application/x-java-class",
    ".com": "text/plain",
    ".conf": "text/plain",
    ".cpio": "application/x-cpio",
    ".cpp": "text/x-c",
    ".cpt": "application/x-cpt",
    ".crl": "application/pkix-crl",
    ".crt": "application/x-x509-user-cert",
    ".csh": "application/x-csh",
    ".css": "text/css",
    ".csv": "text/csv",
    ".cxx": "text/plain",
    ".dcr": "application/x-director",
    ".deepv": "application/x-deepv",
    ".def": "text/plain",
    ".der": "application/x-x509-ca-cert",
    ".dif": "video/x-dv",
    ".dir": "application/x-director",
    ".dl": "video/x-dl",
    ".doc": "application/msword",
    ".docx": "application/vnd.openxmlformats-officedocument.wordprocessingml.document",
    ".dot": "application/msword",
    ".dp": "application/commonground",
    ".drw": "application/drafting",
    ".dump": "application/octet-stream",
    ".dv": "video/x-dv",
    ".dvi": "application/x-dvi",
    ".dwf": "model/vnd.dwf",
    ".dwg": "image/x-dwg",
    ".dxf": "image/x-dwg",
    ".dxr": "application/x-director",
    ".el": "text/x-script.elisp",
    ".elc": "application/x-elc",
    ".env": "application/x-envoy",
    ".eot": "application/vnd.ms-fontobject",
    ".eps": "application/postscript",
    ".es": "application/x-esrehber",
    ".etx": "text/x-setext",
    ".evy": "application/x-envoy",
    ".exe": "application/octet-stream",
    ".f": "text/x-fortran",
    ".f77": "text/x-fortran",
    ".f90": "text/x-fortran",
    ".fdf": "application/vnd.fdf",
    ".fif": "image/fif",
    ".flac": "audio/flac",
    ".fli": "video/x-fli",
    ".flo": "image/florian",
    ".flx": "text/vnd.fmi.flexstor",
    ".fmf": "video/x-atomic3d-feature",
    ".for": "text/x-fortran",
    ".fpx": "image/vnd.net-fpx",
    ".frl": "application/freeloader",
    ".funk": "audio/make",
    ".g": "text/plain",
    ".g3": "image/g3fax",
    ".gif": "image/gif",
    ".gl": "video/x-gl",
    ".gsd": "audio/x-gsm",
    ".gsm": "audio/x-gsm",
    ".gsp": "application/x-gsp",
    ".gss": "application/x-gss",
    ".gtar": "application/x-gtar",
    ".gz": "application/x-gzip",
    ".gzip": "multipart/x-gzip",
    ".h": "text/plain",
    ".hdf": "application/x-hdf",
    ".help": "application/x-helpfile",
    ".hgl": "application/vnd.hp-hpgl",
    ".hh": "text/x-h",
    ".hlb": "text/x-script",
    ".hlp": "application/x-winhelp",
    ".hpg": "application/vnd.hp-hpgl",
    ".hpgl": "application/vnd.hp-hpgl",
    ".hqx": "application/x-mac-binhex40",
    ".hta": "application/hta",
    ".htc": "text/x-component",
    ".htm": "text/html",
    ".html": "text/html",
    ".htmls": "text/html",
    ".htt": "text/webviewhtml",
    ".htx": "text/html",
    ".ice": "x-conference/x-cooltalk",
    ".ico": "image/vnd.microsoft.icon",
    ".ics": "text/calendar",
    ".idc": "text/plain",
    ".ief": "image/ief",
    ".iefs": "image/ief",
    ".iges": "model/iges",
    ".igs": "model/iges",
    ".ima": "application/x-ima",
    ".imap": "application/x-httpd-imap",
    ".inf": "application/inf",
    ".ins": "application/x-internett-signup",
    ".ip": "application/x-ip2",
    ".isu": "video/x-isvideo",
    ".it": "audio/it",
    ".iv": "application/x-inventor",
    ".ivr": "i-world/i-vrml",
    ".ivy": "application/x-livescreen",
    ".jam": "audio/x-jam",
    ".jav": "text/x-java-source",
    ".java": "text/x-java-source",
    ".jcm": "application/x-java-commerce",
    ".jfif": "image/pjpeg",
    ".jfif-bnl": "image/jpeg",
    ".jpe": "image/jpeg",
    ".jpeg": "image/jpeg",
    ".jpg": "image/jpg",
    ".jps": "image/x-jps",
    ".js": "application/javascript",
    ".json": "application/json",
    ".jut": "image/jutvision",
    ".kar": "music/x-karaoke",
    ".ksh": "text/plain",
    ".la": "audio/x-nspaudio",
    ".lam": "audio/x-liveaudio",
    ".latex": "application/x-latex",
    ".lha": "application/x-lha",
    ".lhx": "application/octet-stream",
    ".list": "text/plain",
    ".lma": "audio/x-nspaudio",
    ".log": "text/plain",
    ".lsp": "text/x-script.lisp",
    ".lst": "text/plain",
    ".lsx": "text/x-la-asf",
    ".ltx": "application/x-latex",
    ".lzh": "application/x-lzh",
    ".lzx": "application/x-lzx",
    ".m": "text/x-m",
    ".m1v": "video/mpeg",
    ".m2a": "audio/mpeg",
    ".m2v": "video/mpeg",
    ".m3u": "application/vnd.apple.mpegurl",
    ".man": "application/x-troff-man",
    ".map": "application/x-navimap",
    ".mar": "text/plain",
    ".mbd": "application/mbedlet",
    ".mc": "application/x-magic-cap-package-1.0",
    ".mcd": "application/x-mathcad",
    ".mcf": "text/mcf",
    ".mcp": "application/netmc",
    ".me": "application/x-troff-me",
    ".mht": "message/rfc822",
    ".mhtml": "message/rfc822",
    ".mid": "audio/midi",
    ".midi": "audio/midi",
    ".mif": "application/x-mif",
    ".mime": "www/mime",
    ".mjf": "audio/x-vnd.audioexplosion.mjuicemediafile",
    ".mjpg": "video/x-motion-jpeg",
    ".mka": "audio/x-matroska",
    ".mkv": "video/x-matroska",
    ".mm": "application/x-meme",
    ".mme": "application/base64",
    ".mod": "audio/x-mod",
    ".moov": "video/quicktime",
    ".mov": "video/quicktime",
    ".movie": "video/x-sgi-movie",
    ".mp2": "audio/mpeg",
    ".mp3": "audio/mpeg",
    ".mp4": "video/mp4",
    ".mpa": "video/mpeg",
    ".mpc": "application/x-project",
    ".mpe": "video/mpeg",
    ".mpeg": "video/mpeg",
    ".mpg": "video/mpeg",
    ".mpga": "audio/mpeg",
    ".mpp": "application/vnd.ms-project",
    ".mpt": "application/x-project",
    ".mpv": "application/x-project",
    ".mpx": "application/x-project",
    ".mrc": "application/marc",
    ".ms": "application/x-troff-ms",
    ".mv": "video/x-sgi-movie",
    ".my": "audio/make",
    ".mzz": "application/x-vnd.audioexplosion.mzz",
    ".nap": "image/naplps",
    ".naplps": "image/naplps",
    ".nc": "application/x-netcdf",
    ".ncm": "application/vnd.nokia.configuration-message",
    ".nif": "image/x-niff",
    ".niff": "image/x-niff",
    ".nix": "application/x-mix-transfer",
    ".nsc": "application/x-conference",
    ".nvd": "application/x-navidoc",
    ".o": "application/octet-stream",
    ".oda": "application/oda",
    ".ogg": "video/ogg",
    ".omc": "application/x-omc",
    ".omcd": "application/x-omcdatamaker",
    ".omcr": "application/x-omcregerator",
    ".otf": "font/otf",
    ".p": "text/x-pascal",
    ".p10": "application/x-pkcs10",
    ".p12": "application/x-pkcs12",
    ".p7a": "application/x-pkcs7-signature",
    ".p7c": "application/pkcs7-mime",
    ".p7m": "application/x-pkcs7-mime",
    ".p7r": "application/x-pkcs7-certreqresp",
    ".p7s": "application/pkcs7-signature",
    ".part": "application/pro_eng",
    ".pas": "text/pascal",
    ".pbm": "image/x-portable-bitmap",
    ".pcl": "application/x-pcl",
    ".pct": "image/pict",
    ".pcx": "image/x-pcx",
    ".pdb": "chemical/x-pdb",
    ".pdf": "application/pdf",
    ".pfunk": "audio/make.my.funk",
    ".pgm": "image/x-portable-graymap",
    ".pic": "image/pict",
    ".pict": "image/pict",
    ".pkg": "application/x-newton-compatible-pkg",
    ".pko": "application/vnd.ms-pki.pko",
    ".pl": "text/plain",
    ".plx": "application/x-pixclscript",
    ".pm": "text/x-script.perl-module",
    ".pm4": "application/x-pagemaker",
    ".pm5": "application/x-pagemaker",
    ".png": "image/png",
    ".pnm": "image/x-portable-anymap",
    ".pot": "application/vnd.ms-powerpoint",
    ".pov": "model/x-pov",
    ".ppa": "application/vnd.ms-powerpoint",
    ".ppm": "image/x-portable-pixmap",
    ".pps": "application/vnd.ms-powerpoint",
    ".ppt": "application/vnd.ms-powerpoint",
    ".pptx": "application/vnd.openxmlformats-officedocument.presentationml.presentation",
    ".ppz": "application/mspowerpoint",
    ".pre": "application/x-freelance",
    ".prt": "application/pro_eng",
    ".ps": "application/postscript",
    ".psd": "application/octet-stream",
    ".pvu": "paleovu/x-pv",
    ".pwz": "application/vnd.ms-powerpoint",
    ".py": "text/x-python",
    ".pyc": "application/x-python-code",
    ".qcp": "audio/vnd.qcelp",
    ".qd3": "x-world/x-3dmf",
    ".qd3d": "x-world/x-3dmf",
    ".qif": "image/x-quicktime",
    ".qt": "video/quicktime",
    ".qtc": "video/x-qtc",
    ".qti": "image/x-quicktime",
    ".qtif": "image/x-quicktime",
    ".ra": "audio/x-pn-realaudio",
    ".ram": "application/x-pn-realaudio",
    ".ras": "image/x-cmu-raster",
    ".rast": "image/cmu-raster",
    ".rar": "application/vnd.rar",
    ".rexx": "text/x-script.rexx",
    ".rf": "image/vnd.rn-realflash",
    ".rgb": "image/x-rgb",
    ".rm": "audio/x-pn-realaudio",
    ".rmi": "audio/mid",
    ".rmm": "audio/x-pn-realaudio",
    ".rmp": "audio/x-pn-realaudio-plugin",
    ".rng": "application/vnd.nokia.ringing-tone",
    ".rnx": "application/vnd.rn-realplayer",
    ".roff": "application/x-troff",
    ".rp": "image/vnd.rn-realpix",
    ".rpm": "audio/x-pn-realaudio-plugin",
    ".rt": "text/vnd.rn-realtext",
    ".rtf": "application/rtf",
    ".rtx": "text/richtext",
    ".rv": "video/vnd.rn-realvideo",
    ".s": "text/x-asm",
    ".s3m": "audio/s3m",
    ".saveme": "application/octet-stream",
    ".sbk": "application/x-tbook",
    ".scm": "video/x-scm",
    ".sdml": "text/plain",
    ".sdp": "application/x-sdp",
    ".sdr": "application/sounder",
    ".sea": "application/x-sea",
    ".set": "application/set",
    ".sgm": "text/x-sgml",
    ".sgml": "text/x-sgml",
    ".sh": "application/x-sh",
    ".shar": "application/x-shar",
    ".shtml": "text/x-server-parsed-html",
    ".sid": "audio/x-psid",
    ".sit": "application/x-stuffit",
    ".skd": "application/x-koan",
    ".skm": "application/x-koan",
    ".skp": "application/x-koan",
    ".skt": "application/x-koan",
    ".sl": "application/x-seelogo",
    ".smi": "application/smil",
    ".smil": "application/smil",
    ".snd": "audio/basic",
    ".sol": "application/solids",
    ".spc": "text/x-speech",
    ".spl": "application/futuresplash",
    ".spr": "application/x-sprite",
    ".sprite": "application/x-sprite",
    ".src": "application/x-wais-source",
    ".ssi": "text/x-server-parsed-html",
    ".ssm": "application/streamingmedia",
    ".sst": "application/vnd.ms-pki.certstore",
    ".step": "application/step",
    ".stl": "application/x-navistyle",
    ".stp": "application/step",
    ".sv4cpio": "application/x-sv4cpio",
    ".sv4crc": "application/x-sv4crc",
    ".svf": "image/x-dwg",
    ".svg": "image/svg+xml",
    ".svr": "x-world/x-svr",
    ".swf": "application/x-shockwave-flash",
    ".t": "application/x-troff",
    ".talk": "text/x-speech",
    ".tar": "application/x-tar",
    ".tbk": "application/x-tbook",
    ".tcl": "application/x-tcl",
    ".tcsh": "text/x-script.tcsh",
    ".tex": "application/x-tex",
    ".texi": "application/x-texinfo",
    ".texinfo": "application/x-texinfo",
    ".text": "text/plain",
    ".tgz": "application/x-compressed",
    ".tif": "image/tiff",
    ".tiff": "image/tiff",
    ".tr": "application/x-troff",
    ".ts": "video/mp2t",
    ".tsi": "audio/tsp-audio",
    ".tsp": "audio/tsplayer",
    ".tsv": "text/tab-separated-values",
    ".turbot": "image/florian",
    ".txt": "text/plain",
    ".uil": "text/x-uil",
    ".uni": "text/uri-list",
    ".unis": "text/uri-list",
    ".unv": "application/i-deas",
    ".uri": "text/uri-list",
    ".uris": "text/uri-list",
    ".ustar": "application/x-ustar",
    ".uu": "text/x-uuencode",
    ".uue": "text/x-uuencode",
    ".vcd": "application/x-cdlink",
    ".vcs": "text/x-vcalendar",
    ".vda": "application/vda",
    ".vdo": "video/vdo",
    ".vew": "application/groupwise",
    ".viv": "video/vnd.vivo",
    ".vivo": "video/vnd.vivo",
    ".vmd": "application/vocaltec-media-desc",
    ".vmf": "application/vocaltec-media-file",
    ".voc": "audio/x-voc",
    ".vos": "video/vosaic",
    ".vox": "audio/voxware",
    ".vqe": "audio/x-twinvq-plugin",
    ".vqf": "audio/x-twinvq",
    ".vql": "audio/x-twinvq-plugin",
    ".vrml": "x-world/x-vrml",
    ".vrt": "x-world/x-vrt",
    ".vsd": "application/x-visio",
    ".vst": "application/x-visio",
    ".vsw": "application/x-visio",
    ".w60": "application/wordperfect6.0",
    ".w61": "application/wordperfect6.1",
    ".w6w": "application/msword",
    ".wav": "audio/x-wav",
    ".wb1": "application/x-qpro",
    ".wbmp": "image/vnd.wap.wbmp",
    ".web": "application/vnd.xara",
    ".webm": "video/webm",
    ".webp": "image/webp",
    ".wiz": "application/msword",
    ".wk1": "application/x-123",
    ".wmf": "windows/metafile",
    ".wml": "text/vnd.wap.wml",
    ".wmlc": "application/vnd.wap.wmlc",
    ".wmls": "text/vnd.wap.wmlscript",
    ".wmlsc": "application/vnd.wap.wmlscriptc",
    ".word": "application/msword",
    ".woff": "font/woff",
    ".woff2": "font/woff2",
    ".wp": "application/wordperfect",
    ".wp5": "application/wordperfect6.0",
    ".wp6": "application/wordperfect",
    ".wpd": "application/x-wpwin",
    ".wq1": "application/x-lotus",
    ".wri": "application/x-wri",
    ".wrl": "x-world/x-vrml",
    ".wrz": "x-world/x-vrml",
    ".wsc": "text/scriplet",
    ".wsrc": "application/x-wais-source",
    ".wtk": "application/x-wintalk",
    ".xbm": "image/x-xbitmap",
    ".xdr": "video/x-amt-demorun",
    ".xgz": "xgl/drawing",
    ".xif": "image/vnd.xiff",
    ".xl": "application/excel",
    ".xla": "application/x-msexcel",
    ".xlb": "application/vnd.ms-excel",
    ".xlc": "application/x-excel",
    ".xld": "application/x-excel",
    ".xlk": "application/x-excel",
    ".xll": "application/x-excel",
    ".xlm": "application/x-excel",
    ".xls": "application/vnd.ms-excel",
    ".xlt": "application/x-excel",
    ".xlv": "application/x-excel",
    ".xlw": "application/x-msexcel",
    ".xm": "audio/xm",
    ".xml": "text/xml",
    ".xmz": "xgl/movie",
    ".xpix": "application/x-vnd.ls-xpix",
    ".xpm": "image/x-xpixmap",
    ".x-ng": "image/png",
    ".xlsx": "application/vnd.openxmlformats-officedocument.spreadsheetml.sheet",
    ".xsr": "video/x-amt-showrun",
    ".xwd": "image/x-xwindowdump",
    ".xyz": "chemical/x-pdb",
    ".yaml": "application/x-yaml",
    ".yml": "application/x-yaml",
    ".z": "application/x-compressed",
    ".zip": "application/zip",
    ".zoo": "application/octet-stream",
    ".zsh": "text/x-script.zsh",
    ".mjs": "application/javascript",
    ".webmanifest": "application/manifest+json",
    ".dll": "application/octet-stream",
    ".obj": "application/octet-stream",
    ".so": "application/octet-stream",
    ".m3u8": "application/vnd.apple.mpegurl",
    ".wasm": "application/wasm",
    ".h5": "application/x-hdf5",
    ".pfx": "application/x-pkcs12",
    ".pyo": "application/x-python-code",
    ".xsl": "application/xml",
    ".rdf": "application/xml",
    ".wsdl": "application/xml",
    ".xpdl": "application/xml",
    ".3gp": "audio/3gpp",
    ".3gpp": "audio/3gpp",
    ".3g2": "audio/3gpp2",
    ".3gpp2": "audio/3gpp2",
    ".aac": "audio/aac",
    ".adts": "audio/aac",
    ".loas": "audio/aac",
    ".ass": "audio/aac",
    ".opus": "audio/opus",
    ".heic": "image/heic",
    ".heif": "image/heif",
    ".eml": "message/rfc822",
    ".nws": "message/rfc822",
    ".bat": "text/plain",
    ".vcf": "text/x-vcard",
    ".xul": "text/xul",
}


def guess(file: t.Union[str, Path]) -> str:
    return _types.get(Path(file).suffix, _default)



from __future__ import annotations

import typing as t

from composio import exceptions
from composio.client import HttpClient
from composio.client.types import (
    AuthSchemeL,
    toolkit_list_params,
    toolkit_list_response,
    toolkit_retrieve_response,
)
from composio.core.models.connected_accounts import ConnectedAccounts

from .base import Resource

AuthFieldsT: t.TypeAlias = t.List[
    toolkit_retrieve_response.AuthConfigDetailFieldsConnectedAccountInitiationRequired
    | toolkit_retrieve_response.AuthConfigDetailFieldsConnectedAccountInitiationOptional
    | toolkit_retrieve_response.AuthConfigDetailFieldsAuthConfigCreationRequired
    | toolkit_retrieve_response.AuthConfigDetailFieldsAuthConfigCreationOptional
]


class Toolkits(Resource):
    """
    Toolkits are a collectiono of tools that can be used to perform various tasks.
    They're conceptualized as a set of tools. Ex: Github toolkit can perform
    Github actions via its collection of tools. This is a replacement of the
    `apps` concept in the earlier versions of the SDK.
    """

    connected_accounts: ConnectedAccounts

    def __init__(self, client: HttpClient):
        super().__init__(client)
        self.connected_accounts = ConnectedAccounts(client)

    @t.overload
    def get(self) -> list[toolkit_list_response.Item]:
        """Get all toolkits."""

    @t.overload
    def get(self, slug: str) -> toolkit_retrieve_response.ToolkitRetrieveResponse:
        """Get a toolkit by slug."""

    @t.overload
    def get(
        self,
        *,
        query: toolkit_list_params.ToolkitListParams,
    ) -> list[toolkit_list_response.Item]:
        """Get a list of toolkits by query."""

    def get(
        self,
        slug: t.Optional[str] = None,
        *,
        query: t.Optional[toolkit_list_params.ToolkitListParams] = None,
    ) -> t.Union[
        toolkit_retrieve_response.ToolkitRetrieveResponse,
        list[toolkit_list_response.Item],
    ]:
        if slug is not None:
            return self._client.toolkits.retrieve(slug=slug)
        return self._client.toolkits.list(**(query or {})).items

    def list_categories(self):
        """List all categories of toolkits."""
        return self._client.toolkits.retrieve_categories().items

    def _get_auth_config_id(self, toolkit: str) -> str:
        """Get the auth config ID for a toolkit."""
        auth_configs = self._client.auth_configs.list(toolkit_slug=toolkit)
        if len(auth_configs.items) > 0:
            (auth_config, *_) = sorted(
                auth_configs.items,
                key=lambda x: t.cast(str, x.created_at),
                reverse=True,
            )
            return auth_config.id

        return self._client.auth_configs.create(
            toolkit={"slug": toolkit},
            auth_config={
                "type": "use_composio_managed_auth",
                "restrict_to_following_tools": [toolkit],
            },
        ).auth_config.id

    def authorize(self, *, user_id: str, toolkit: str):
        """
        Authorize a user to a toolkit

        If auth config is not found, it will be created using composio managed auth.

        :param user_id: The ID of the user to authorize.
        :param toolkit: The slug of the toolkit to authorize.
        :return: The connection request.
        """
        return self.connected_accounts.initiate(
            user_id=user_id,
            auth_config_id=self._get_auth_config_id(
                toolkit=toolkit,
            ),
        )

    def get_connected_account_initiation_fields(
        self,
        toolkit: str,
        auth_scheme: AuthSchemeL,
        required_only: bool = False,
    ) -> AuthFieldsT:
        """
        Get the required property for a given toolkit and auth scheme.
        """
        details = self._client.toolkits.retrieve(slug=toolkit).auth_config_details or []
        for auth_detail in details:
            if auth_detail.mode != auth_scheme:
                continue

            if required_only:
                return t.cast(
                    AuthFieldsT,
                    auth_detail.fields.connected_account_initiation.required,
                )

            return t.cast(
                AuthFieldsT,
                auth_detail.fields.connected_account_initiation.required
                + auth_detail.fields.connected_account_initiation.optional,
            )

        raise exceptions.InvalidParams(
            f"auth config details not found with {toolkit=} and {auth_scheme=}"
        )

    def get_auth_config_creation_fields(
        self,
        toolkit: str,
        auth_scheme: AuthSchemeL,
        required_only: bool = False,
    ) -> AuthFieldsT:
        """
        Get the required property for a given toolkit and auth scheme.
        """
        info = self._client.toolkits.retrieve(slug=toolkit)
        for auth_detail in info.auth_config_details or []:
            if auth_detail.mode != auth_scheme:
                continue

            if required_only:
                return t.cast(
                    AuthFieldsT,
                    auth_detail.fields.auth_config_creation.required,
                )

            return t.cast(
                AuthFieldsT,
                auth_detail.fields.auth_config_creation.required
                + auth_detail.fields.auth_config_creation.optional,
            )

        raise exceptions.InvalidParams(
            f"auth config details not found with {toolkit=} and {auth_scheme=}"
        )



from composio import Composio

composio = Composio()

# List all toolkits
toolkits = composio.toolkits.get()
print(toolkits)

# Get a toolkit by slug
toolkit = composio.toolkits.get(slug="github")
print(toolkit)

# List all toolkits categories
categories = composio.toolkits.list_categories()
print(categories)

# Authorize a user to a toolkit
connection_request = composio.toolkits.authorize(user_id="123", toolkit="github")
print(connection_request)



from __future__ import annotations

import functools
import time
import typing as t

from composio import exceptions
from composio.client import HttpClient
from composio.client.types import (
    connected_account_create_params,
    connected_account_retrieve_response,
    connected_account_update_status_response,
)

from .base import Resource


class ConnectionRequest(Resource):
    """
    A connection request.

    This class is used to manage connection requests.
    """

    DEFAULT_WAIT_TIMEOUT = 60.0  # Seconds

    def __init__(
        self,
        id: str,
        status: str,
        redirect_url: t.Optional[str],
        client: HttpClient,
    ):
        """
        Initialize the connection request.

        :param id: The ID of the connection request.
        :param status: The status of the connection request.
        :param redirect_url: The redirect URL of the connection request.
        :param client: The client to use for the connection request.
        """
        super().__init__(client)
        self.id = id
        self.status = status
        self.redirect_url = redirect_url

    def wait_for_connection(
        self,
        timeout: t.Optional[float] = None,
    ) -> connected_account_retrieve_response.ConnectedAccountRetrieveResponse:
        """
        Wait for the connection to be established.

        :param timeout: The timeout to wait for the connection to be established.
        :return: Connected account object.
        """
        timeout = self.DEFAULT_WAIT_TIMEOUT if timeout is None else timeout
        deadline = time.time() + timeout
        while deadline > time.time():
            connection = self._client.connected_accounts.retrieve(nanoid=self.id)
            self.status = connection.status
            if self.status != "ACTIVE":
                time.sleep(1)
                continue
            return connection

        raise exceptions.ComposioSDKTimeoutError(
            message=f"Timeout while waiting for connection {self.id} to be active",
        )


class AuthScheme:
    """
    Collection of auth scheme helpers.
    """

    def oauth1(
        self, options: connected_account_create_params.ConnectionStateUnionMember0Val
    ) -> connected_account_create_params.ConnectionState:
        """
        Create a new connected account using OAuth 1.0.
        """
        return {
            "auth_scheme": "OAUTH1",
            "val": t.cast(
                connected_account_create_params.ConnectionStateUnionMember0Val,
                {
                    **options,
                    "status": "INITIALISING",
                },
            ),
        }

    def oauth2(
        self, options: connected_account_create_params.ConnectionStateUnionMember1Val
    ) -> connected_account_create_params.ConnectionState:
        """
        Create a new connected account using OAuth 1.0.
        """
        return {
            "auth_scheme": "OAUTH2",
            "val": t.cast(
                connected_account_create_params.ConnectionStateUnionMember1Val,
                {
                    **options,
                    "status": "INITIALISING",
                },
            ),
        }

    def composio_link(
        self, options: connected_account_create_params.ConnectionStateUnionMember2Val
    ) -> connected_account_create_params.ConnectionState:
        """
        Create a new connected account using Composio Link.
        """
        return {
            "auth_scheme": "COMPOSIO_LINK",
            "val": t.cast(
                connected_account_create_params.ConnectionStateUnionMember2Val,
                {
                    **options,
                    "status": "INITIALIZING",
                },
            ),
        }

    def api_key(
        self, options: connected_account_create_params.ConnectionStateUnionMember3Val
    ) -> connected_account_create_params.ConnectionState:
        """
        Create a new connected account using an API key.
        """
        return {
            "auth_scheme": "API_KEY",
            "val": t.cast(
                connected_account_create_params.ConnectionStateUnionMember3Val,
                {
                    **options,
                    "status": "ACTIVE",
                },
            ),
        }

    def basic(
        self, options: connected_account_create_params.ConnectionStateUnionMember4Val
    ) -> connected_account_create_params.ConnectionState:
        """
        Create a new connected account using basic auth.
        """
        return {
            "auth_scheme": "BASIC",
            "val": t.cast(
                connected_account_create_params.ConnectionStateUnionMember4Val,
                {
                    **options,
                    "status": "ACTIVE",
                },
            ),
        }

    def bearer_token(
        self, options: connected_account_create_params.ConnectionStateUnionMember5Val
    ) -> connected_account_create_params.ConnectionState:
        """
        Create a new connected account using a bearer token.
        """
        return {
            "auth_scheme": "BEARER_TOKEN",
            "val": t.cast(
                connected_account_create_params.ConnectionStateUnionMember5Val,
                {
                    **options,
                    "status": "ACTIVE",
                },
            ),
        }

    def google_service_account(
        self, options: connected_account_create_params.ConnectionStateUnionMember6Val
    ) -> connected_account_create_params.ConnectionState:
        """
        Create a new connected account using a Google service account.
        """
        return {
            "auth_scheme": "GOOGLE_SERVICE_ACCOUNT",
            "val": t.cast(
                connected_account_create_params.ConnectionStateUnionMember6Val,
                {
                    **options,
                    "status": "ACTIVE",
                },
            ),
        }

    def no_auth(
        self, options: connected_account_create_params.ConnectionStateUnionMember7Val
    ) -> connected_account_create_params.ConnectionState:
        """
        Create a new connected account using no auth.
        """
        return {
            "auth_scheme": "NO_AUTH",
            "val": t.cast(
                connected_account_create_params.ConnectionStateUnionMember7Val,
                {
                    **options,
                    "status": "ACTIVE",
                },
            ),
        }

    def calcom_auth(
        self, options: connected_account_create_params.ConnectionStateUnionMember8Val
    ) -> connected_account_create_params.ConnectionState:
        """
        Create a new connected account using Cal.com auth.
        """
        return {
            "auth_scheme": "CALCOM_AUTH",
            "val": t.cast(
                connected_account_create_params.ConnectionStateUnionMember8Val,
                {
                    **options,
                    "status": "ACTIVE",
                },
            ),
        }

    def billcom_auth(
        self, options: connected_account_create_params.ConnectionStateUnionMember9Val
    ) -> connected_account_create_params.ConnectionState:
        """
        Create a new connected account using Bill.com auth.
        """
        return {
            "auth_scheme": "BILLCOM_AUTH",
            "val": t.cast(
                connected_account_create_params.ConnectionStateUnionMember9Val,
                {
                    **options,
                    "status": "ACTIVE",
                },
            ),
        }

    def basic_with_jwt(
        self, options: connected_account_create_params.ConnectionStateUnionMember10Val
    ) -> connected_account_create_params.ConnectionState:
        """
        Create a new connected account using basic auth with JWT.
        """
        return {
            "auth_scheme": "BASIC_WITH_JWT",
            "val": t.cast(
                connected_account_create_params.ConnectionStateUnionMember10Val,
                {
                    **options,
                    "status": "ACTIVE",
                },
            ),
        }


class ConnectedAccounts:
    """
    Manage connected accounts.

    This class is used to manage connected accounts in the Composio SDK.
    These are used to authenticate with third-party services.
    """

    enable: t.Callable[
        [str],
        connected_account_update_status_response.ConnectedAccountUpdateStatusResponse,
    ]
    """Enable a connected account."""

    disable: t.Callable[
        [str],
        connected_account_update_status_response.ConnectedAccountUpdateStatusResponse,
    ]
    """Disable a connected account."""

    def __init__(self, client: HttpClient):
        """
        Initialize the connected accounts resource.

        :param client: The client to use for the connected accounts resource.
        """
        self._client = client
        self.get = self._client.connected_accounts.retrieve
        self.list = self._client.connected_accounts.list
        self.delete = self._client.connected_accounts.delete
        self.update_status = self._client.connected_accounts.update_status
        self.enable = functools.partial(
            self._client.connected_accounts.update_status,
            enabled=True,
        )
        self.disable = functools.partial(
            self._client.connected_accounts.update_status,
            enabled=False,
        )

    def initiate(
        self,
        user_id: str,
        auth_config_id: str,
        *,
        callback_url: t.Optional[str] = None,
        config: t.Optional[connected_account_create_params.ConnectionState] = None,
    ) -> ConnectionRequest:
        """
        Compound function to create a new coneected account. This function creates
        a new connected account and returns a connection request.

        Users can then wait for the connection to be established using the
        `wait_for_connection` method.

        :param user_id: The user ID to create the connected account for.
        :param auth_config_id: The auth config ID to create the connected account for.
        :param callback_url: Callback URL to use for OAuth apps.
        :param options: The options to create the connected account with.
        :return: The connection request.
        """
        connection: dict[str, t.Any] = {"user_id": user_id}
        if callback_url is not None:
            connection["callback_url"] = connection

        if config is not None:
            connection["state"] = connection

        response = self._client.connected_accounts.create(
            auth_config={"id": auth_config_id},
            connection=t.cast(connected_account_create_params.Connection, connection),
        )
        return ConnectionRequest(
            id=response.id,
            status=response.connection_data.val.status,
            redirect_url=getattr(response.connection_data.val, "redirect_url", None),
            client=self._client,
        )


auth_scheme = AuthScheme()



import typing as t

from .agentic import AgenticProvider, AgenticProviderExecuteFn
from .none_agentic import NonAgenticProvider, NoneAgenticProviderExecuteFn

TProvider = t.TypeVar("TProvider", bound=AgenticProvider | NonAgenticProvider)

__all__ = [
    "TProvider",
    "AgenticProvider",
    "NonAgenticProvider",
    "NoneAgenticProviderExecuteFn",
    "AgenticProviderExecuteFn",
]



"""
This module is a light wrapper around the auto-generated composio client types.
"""

import typing as t

from composio_client import NotGiven
from composio_client.types import (
    auth_config_create_params,
    auth_config_create_response,
    auth_config_list_params,
    auth_config_list_response,
    auth_config_retrieve_response,
    auth_config_update_params,
    connected_account_create_params,
    connected_account_create_response,
    connected_account_list_params,
    connected_account_list_response,
    connected_account_retrieve_response,
    connected_account_update_status_response,
    tool_execute_params,
    tool_execute_response,
    tool_list_response,
    tool_proxy_params,
    tool_proxy_response,
    toolkit_list_params,
    toolkit_list_response,
    toolkit_retrieve_response,
    trigger_instance_upsert_response,
)

Tool: t.TypeAlias = tool_list_response.Item
ToolkitMinimal: t.TypeAlias = tool_list_response.ItemToolkit
AuthConfig: t.TypeAlias = connected_account_create_params.AuthConfig

Oauth1L: t.TypeAlias = t.Literal["OAUTH1"]
Oauth2L: t.TypeAlias = t.Literal["OAUTH2"]
ApiKeyL: t.TypeAlias = t.Literal["API_KEY"]
BasicL: t.TypeAlias = t.Literal["BASIC"]
NoAuthL: t.TypeAlias = t.Literal["NO_AUTH"]
SnowflakeL: t.TypeAlias = t.Literal["SNOWFLAKE"]
CalcomAuthL: t.TypeAlias = t.Literal["CALCOM_AUTH"]
BearerTokenL: t.TypeAlias = t.Literal["BEARER_TOKEN"]
BillcomAuthL: t.TypeAlias = t.Literal["BILLCOM_AUTH"]
ComposioLinkL: t.TypeAlias = t.Literal["COMPOSIO_LINK"]
BasicWithJwtL: t.TypeAlias = t.Literal["BASIC_WITH_JWT"]
GoogleServiceAccountL: t.TypeAlias = t.Literal["GOOGLE_SERVICE_ACCOUNT"]

AuthSchemeL: t.TypeAlias = t.Literal[
    Oauth1L,
    Oauth2L,
    ApiKeyL,
    BasicL,
    NoAuthL,
    SnowflakeL,
    CalcomAuthL,
    BearerTokenL,
    BillcomAuthL,
    ComposioLinkL,
    BasicWithJwtL,
    GoogleServiceAccountL,
]

__all__ = (
    "auth_config_create_params",
    "auth_config_create_response",
    "auth_config_list_params",
    "auth_config_list_response",
    "auth_config_retrieve_response",
    "auth_config_update_params",
    "connected_account_create_params",
    "connected_account_create_response",
    "connected_account_list_params",
    "connected_account_list_response",
    "connected_account_retrieve_response",
    "connected_account_update_status_response",
    "trigger_instance_upsert_response",
    "tool_execute_params",
    "tool_execute_response",
    "tool_list_response",
    "tool_proxy_params",
    "tool_proxy_response",
    "toolkit_list_params",
    "toolkit_list_response",
    "toolkit_retrieve_response",
    "Tool",
    "ToolkitMinimal",
    "AuthConfig",
    "NotGiven",
    "AuthSchemeL",
)



from __future__ import annotations

import os
import typing as t

import typing_extensions as te

from composio import exceptions
from composio.client import DEFAULT_MAX_RETRIES, APIEnvironment, HttpClient
from composio.core.models import (
    AuthConfigs,
    ConnectedAccounts,
    Toolkits,
    Tools,
    Triggers,
)
from composio.core.models.base import allow_tracking
from composio.core.provider import TProvider
from composio.core.provider._openai import OpenAIProvider
from composio.utils.logging import WithLogger

_DEFAULT_PROVIDER = OpenAIProvider()


class SDKConfig(te.TypedDict):
    environment: te.NotRequired[APIEnvironment]
    api_key: te.NotRequired[str]
    base_url: te.NotRequired[str]
    timeout: te.NotRequired[int]
    max_retries: te.NotRequired[int]
    allow_tracking: te.NotRequired[bool]


class Composio(t.Generic[TProvider], WithLogger):
    """
    Composio SDK for Python.
    """

    tools: Tools[TProvider]
    provider: TProvider

    def __init__(
        self,
        provider: TProvider = _DEFAULT_PROVIDER,  # type: ignore
        **kwargs: te.Unpack[SDKConfig],
    ) -> None:
        """
        Initialize the Composio SDK.

        :param provider: The provider to use for the SDK.
        :param environment: The environment to use for the SDK.
        :param api_key: The API key to use for the SDK.
        :param base_url: The base URL to use for the SDK.
        :param timeout: The timeout to use for the SDK.
        :param max_retries: The maximum number of retries to use for the SDK.
        """
        WithLogger.__init__(self)
        api_key = kwargs.get("api_key", os.environ.get("COMPOSIO_API_KEY"))
        if not api_key:
            raise exceptions.ApiKeyNotProvidedError()

        allow_tracking.set(kwargs.get("allow_tracking", True))
        self._client = HttpClient(
            environment=kwargs.get("environment", "production"),
            provider=provider.name,
            api_key=api_key,
            base_url=kwargs.get("base_url"),
            timeout=kwargs.get("timeout"),
            max_retries=kwargs.get("max_retries", DEFAULT_MAX_RETRIES),
        )
        self.provider = provider
        self.tools = Tools(client=self._client, provider=self.provider)
        self.toolkits = Toolkits(client=self._client)
        self.triggers = Triggers(client=self._client)
        self.auth_configs = AuthConfigs(client=self._client)
        self.connected_accounts = ConnectedAccounts(client=self._client)

    @property
    def client(self):
        return self._client



import typing as t

import pydantic
import pydantic.error_wrappers
import typing_extensions as te
from crewai.tools import BaseTool


from composio.core.provider import AgenticProvider, AgenticProviderExecuteFn
from composio.types import Tool
from composio.utils.pydantic import parse_pydantic_error
from composio.utils.shared import json_schema_to_model


class CrewAIProvider(AgenticProvider[BaseTool, list[BaseTool]], name="crewai"):
    """
    Composio toolset for CrewiAI framework.
    """

    def wrap_tool(
        self,
        tool: Tool,
        execute_tool: AgenticProviderExecuteFn,
    ) -> BaseTool:
        """Wrap a tool as a CrewAI tool."""

        class Wrapper(BaseTool):

            def _run(self, **kwargs):
                try:
                    return execute_tool(slug=tool.slug, arguments=kwargs)
                except pydantic.ValidationError as e:
                    return {
                        "successful": False,
                        "error": parse_pydantic_error(e),
                        "data": None,
                    }

        return Wrapper(
            name=tool.slug,
            description=tool.description,
            args_schema=json_schema_to_model(
                json_schema=tool.input_parameters,
                skip_default=True,
            ),
        )

    def wrap_tools(
        self,
        tools: t.Sequence[Tool],
        execute_tool: AgenticProviderExecuteFn,
    ) -> list[BaseTool]:
        """Wrap a list of tools as a list of CrewAI tools."""
        return [self.wrap_tool(tool, execute_tool) for tool in tools]



from composio_client import BaseModel

from composio.core.models.base import Resource

INTERNAL_SDK_REALTIME_CREDENTIALS_ENDPOINT = "/api/v3/internal/sdk/realtime/credentials"


class SDKRealtimeCredentialsResponse(BaseModel):
    pusher_key: str
    project_id: str
    pusher_cluster: str


class Internal(Resource):
    """
    Internal resource for getting internal SDK realtime credentials.
    """

    def get_sdk_realtime_credentials(self) -> SDKRealtimeCredentialsResponse:
        """
        Get the SDK realtime credentials.

        :return: The SDK realtime credentials.
        """
        return self._client.get(
            path=INTERNAL_SDK_REALTIME_CREDENTIALS_ENDPOINT,
            cast_to=SDKRealtimeCredentialsResponse,
        )



from fastapi import FastAPI
from fastapi.responses import RedirectResponse

from composio import Composio

# Create a FastAPI app
app = FastAPI()

# Create a Composio client
composio = Composio()


@app.get("/authorize/{toolkit}")
def authorize_app(toolkit: str):
    # retrieve the user id from your app
    user_id = ""

    # retrieve the auth config id from your app
    auth_config_id = ""

    # initiate the connection request
    connection_request = composio.connected_accounts.initiate(
        user_id=user_id,
        auth_config_id=auth_config_id,
    )
    return RedirectResponse(url=connection_request.redirect_url)



import pydantic


def parse_pydantic_error(e: pydantic.ValidationError) -> str:
    """Parse pydantic validation error."""
    message = "Invalid request data provided"
    missing = []
    others = [""]
    for error in e.errors():
        param = ".".join(map(str, error["loc"]))
        if error["type"] == "missing":
            missing.append(param)
            continue
        others.append(error["msg"] + f" on parameter `{param}`")
    if len(missing) > 0:
        message += f"\n- Following fields are missing: {set(missing)}"
    message += "\n- ".join(others)
    return message



import os

from composio import Composio, after_execute, before_execute, schema_modifier
from composio.types import Tool, ToolExecuteParams, ToolExecutionResponse

composio = Composio()


@before_execute(tools=["HACKERNEWS_GET_USER"])
def before_execute_modifier(
    tool: str,
    toolkit: str,
    params: ToolExecuteParams,
) -> ToolExecuteParams:
    # Perform modifications on the request
    print("before_execute_modifier", tool, toolkit)
    return params


@after_execute(tools=["HACKERNEWS_GET_USER"])
def after_execute_modifier(
    tool: str,
    toolkit: str,
    response: ToolExecutionResponse,
) -> ToolExecutionResponse:
    return {
        **response,
        "data": {
            "karama": response["data"]["karama"],
        },
    }


# execute tool
response = composio.tools.execute(
    user_id="default",
    slug="HACKERNEWS_GET_USER",
    arguments={"username": "pg"},
    modifiers=[
        before_execute_modifier,
        after_execute_modifier,
    ],
)
print(response)


@schema_modifier(tools=["HACKERNEWS_GET_USER"])
def modify_schema(
    tool: str,
    toolkit: str,
    schema: Tool,
) -> Tool:
    # Perform modifications on the schema
    print("modify_schema", tool, toolkit)
    return schema


tools = composio.tools.get(
    user_id="default",
    slug="HACKERNEWS_GET_USER",
    modifiers=[modify_schema],
)
print(tools)


@before_execute(toolkits=["NOTION"])
def add_custom_auth(
    tool: str,
    toolkit: str,
    params: ToolExecuteParams,
) -> ToolExecuteParams:
    if params["custom_auth_params"] is None:
        params["custom_auth_params"] = {"parameters": []}

    params["custom_auth_params"]["parameters"].append(  # type: ignore
        {
            "name": "x-api-key",
            "value": os.getenv("NOTION_API_KEY"),
            "in": "header",
        }
    )
    return params


result = composio.tools.execute(
    user_id="default",
    slug="NOTION_GET_DATABASES",
    arguments={},
    modifiers=[
        add_custom_auth,
    ],
)
print(result)



"""
OpenAI provider implementation.
"""

from __future__ import annotations

import json
import time
import typing as t

from openai import Client
from openai.types.beta.thread import Thread
from openai.types.beta.threads.run import Run
from openai.types.chat.chat_completion import ChatCompletion
from openai.types.chat.chat_completion_message_tool_call import (
    ChatCompletionMessageToolCall,
)
from openai.types.chat.chat_completion_tool_param import ChatCompletionToolParam
from openai.types.shared_params.function_definition import FunctionDefinition
from openai.types.shared_params.function_parameters import FunctionParameters

from composio.core.provider import NonAgenticProvider
from composio.types import Modifiers, Tool, ToolExecutionResponse

OpenAITool: t.TypeAlias = ChatCompletionToolParam
OpenAIToolCollection: t.TypeAlias = t.List[OpenAITool]


class OpenAIProvider(
    NonAgenticProvider[OpenAITool, OpenAIToolCollection], name="openai"
):
    """OpenAIProvider class definition"""

    def wrap_tool(self, tool: Tool) -> OpenAITool:
        return ChatCompletionToolParam(
            function=FunctionDefinition(
                name=tool.slug,
                description=tool.description,
                parameters=t.cast(FunctionParameters, tool.input_parameters),
                strict=None,
            ),
            type="function",
        )

    def wrap_tools(self, tools: t.Sequence[Tool]) -> OpenAIToolCollection:
        return [self.wrap_tool(tool) for tool in tools]

    def execute_tool_call(
        self,
        user_id: str,
        tool_call: ChatCompletionMessageToolCall,
        modifiers: t.Optional[Modifiers] = None,
    ) -> ToolExecutionResponse:
        """Execute a tool call.

        :param tool_call: Tool call metadata.
        :param user_id: User ID to use for executing the function call.
        :return: Object containing output data from the tool call.
        """
        return self.execute_tool(
            slug=tool_call.function.name,
            arguments=json.loads(tool_call.function.arguments),
            modifiers=modifiers,
            user_id=user_id,
        )

    def handle_tool_calls(
        self,
        user_id: str,
        response: ChatCompletion,
        modifiers: t.Optional[Modifiers] = None,
    ) -> t.List[ToolExecutionResponse]:
        """
        Handle tool calls from OpenAI chat completion object.

        :param response: Chat completion object from
                        openai.OpenAI.chat.completions.create function call
        :param entity_id: Entity ID to use for executing the function call.
        :return: A list of output objects from the function calls.
        """
        outputs = []
        for choice in response.choices:
            if choice.message.tool_calls is None:
                continue

            for tool_call in choice.message.tool_calls:
                outputs.append(
                    self.execute_tool_call(
                        user_id=user_id,
                        tool_call=tool_call,
                        modifiers=modifiers,
                    )
                )
        return outputs

    def handle_assistant_tool_calls(
        self,
        user_id: str,
        run: Run,
    ) -> t.List:
        """Wait and handle assistant function calls"""
        tool_outputs: list[dict] = []
        if run.required_action is None:
            return tool_outputs

        for tool_call in run.required_action.submit_tool_outputs.tool_calls:
            tool_outputs.append(
                {
                    "tool_call_id": tool_call.id,
                    "output": json.dumps(
                        self.execute_tool_call(
                            tool_call=t.cast(ChatCompletionMessageToolCall, tool_call),
                            user_id=user_id,
                        )
                    ),
                }
            )
        return tool_outputs

    def wait_and_handle_assistant_tool_calls(
        self,
        user_id: str,
        client: Client,
        run: Run,
        thread: Thread,
    ) -> Run:
        """Wait and handle assistant function calls"""
        while run.status in ("queued", "in_progress", "requires_action"):
            if run.status != "requires_action":
                run = client.beta.threads.runs.retrieve(
                    thread_id=thread.id,
                    run_id=run.id,
                )
                time.sleep(0.5)
                continue

            run = client.beta.threads.runs.submit_tool_outputs(
                thread_id=thread.id,
                run_id=run.id,
                tool_outputs=self.handle_assistant_tool_calls(
                    run=run,
                    user_id=user_id,
                ),
            )
        return run



from datetime import datetime

from openai import OpenAI

from composio import Composio
from composio_openai import OpenAIProvider


# Initialize tools.
openai_client = OpenAI()
composio = Composio(provider=OpenAIProvider())

# Retrieve actions
actions = composio.tools.get(
    user_id="default",
    tools=["NOTION_ADD_PAGE_CONTENT"],
)

# Setup openai assistant
assistant_instruction = (
    "You are a super intelligent personal assistant."
    + "You have been given a set of tools that you are supposed to choose from."
    + "You decide the right tool and execute it."
)
# Prepare assistant
assistant = openai_client.beta.assistants.create(
    name="Personal Assistant",
    instructions=assistant_instruction,
    model="gpt-4-turbo-preview",
    tools=actions,  # type: ignore
)

# Give a task to execute via Openai Assistants
my_task = (
    f"Can you copy all the events in coming week from google calendar to notion? "
    f"Today's date is {datetime.now()} and day is {datetime.now().strftime('%A')}"
)

# create a thread
thread = openai_client.beta.threads.create()
print("Thread ID: ", thread.id)
print("Assistant ID: ", assistant.id)

# start the asssitant with my task
message = openai_client.beta.threads.messages.create(
    thread_id=thread.id,
    role="user",
    content=my_task,
)

# Execute Agent with integrations
run = openai_client.beta.threads.runs.create(
    thread_id=thread.id,
    assistant_id=assistant.id,
)

# Execute function calls
run_after_tool_calls = composio.wait_and_handle_assistant_tool_calls(
    client=openai_client,
    run=run,
    thread=thread,
)

print(run_after_tool_calls)



import asyncio

from agents import Agent, Runner

from composio import Composio
from composio_openai_agents import OpenAIAgentsProvider

# Initialize Composio toolset
composio = Composio(provider=OpenAIAgentsProvider())

# Get all the tools
tools = composio.tools.get(
    user_id="default",
    tools=["GITHUB_STAR_A_REPOSITORY_FOR_THE_AUTHENTICATED_USER"],
)

# Create an agent with the tools
agent = Agent(
    name="GitHub Agent",
    instructions="You are a helpful assistant that helps users with GitHub tasks.",
    tools=tools,
)

# Run the agent
async def main():
    result = await Runner.run(
        starting_agent=agent,
        input=(
            "Star the repository composiohq/composio on GitHub. If done "
            "successfully, respond with 'Action executed successfully'"
        ),
    )
    print(result.final_output)


asyncio.run(main())



import atexit
import functools
import queue as q
import threading as tr
import time
import typing as t

import httpx
import typing_extensions as te

TELEMETRY_URL = "https://telemetry.composio.dev/v1"
METRIC_ENDPOINT = f"{TELEMETRY_URL}/metrics/invocations"
ERROR_ENDPOINT = f"{TELEMETRY_URL}/errors"


class ErrorData(te.TypedDict):
    name: str
    "The name of the error"

    code: te.NotRequired[str]
    "The code of the error"

    errorId: te.NotRequired[str]
    "The error ID of the error"

    message: te.NotRequired[str]
    "The message of the error"

    stack: te.NotRequired[str]
    "The stack trace of the error"


class SourceData(te.TypedDict):
    host: te.NotRequired[str]
    "The name of the source/host"

    service: te.NotRequired[te.Literal["sdk", "apollo", "hermes", "thermos"]]
    "The service of the source"

    language: te.NotRequired[te.Literal["python", "typescript", "go", "rust"]]
    "The language of the function that was invoked"

    version: te.NotRequired[str]
    "The version of the source"

    platform: te.NotRequired[str]
    "The platform of the source"

    environment: te.NotRequired[
        te.Literal["development", "production", "ci", "staging", "test"]
    ]
    "The environment of the source, eg: development, production, ci etc"


class Metadata(te.TypedDict):
    projectId: te.NotRequired[str]
    "The project ID of the source"

    provider: te.NotRequired[str]
    "The provider used in the source"


class TelemetryData(te.TypedDict):
    functionName: str
    "The name of the function that was invoked"

    durationMs: te.NotRequired[float]
    "The duration of the function invocation in milliseconds"

    timestamp: te.NotRequired[float]
    "The timestamp of the function invocation in epoch seconds"

    props: te.NotRequired[t.Dict]
    "The properties of the function invocation"

    source: te.NotRequired[SourceData]
    """Source of the metric"""

    metadata: te.NotRequired[Metadata]
    """Runtime metadata"""

    error: te.NotRequired[ErrorData]
    """Error data."""


EventType: t.TypeAlias = t.Literal["metric", "error"]
Event = t.Tuple[EventType, TelemetryData]
EventQueue: t.TypeAlias = q.Queue[Event]

_queue: t.Optional[EventQueue] = None
_event: t.Optional[tr.Event] = None
_thread: t.Optional[tr.Thread] = None


def _setup():
    global _queue, _event, _thread
    if _queue is None:
        _queue = q.Queue[Event]()

    if _event is None:
        _event = tr.Event()

    if _thread is None:
        _thread = tr.Thread(
            target=_thread_loop,
            kwargs={
                "queue": _queue,
                "event": _event,
            },
            daemon=True,
        )
        _thread.start()
        atexit.register(
            functools.partial(
                _teardown,
                queue=_queue,
                event=_event,
                thread=_thread,
            )
        )

    return _queue, _event, _thread


def _teardown(queue: EventQueue, event: tr.Event, thread: tr.Thread):
    while queue.qsize():
        time.sleep(0.1)

    event.set()
    thread.join()


def _push(event: Event):
    _ = (
        httpx.post(
            url=METRIC_ENDPOINT,
            json=[event[1]],
        )
        if event[0] == "metric"
        else httpx.post(
            url=ERROR_ENDPOINT,
            json=event[1],
        )
    )


def _thread_loop(queue: EventQueue, event: tr.Event):
    while not event.is_set():
        try:
            _push(queue.get(timeout=0.1))
        except q.Empty:
            continue


def push_event(event: Event):
    q, _, _ = _setup()
    q.put(event)


def create_event(type: EventType, **payload: te.Unpack[TelemetryData]) -> Event:
    return type, payload



import typing as t
from enum import Enum
from itertools import chain
from pathlib import Path

import click
import tomli
from semver import VersionInfo

CWD = Path.cwd()
SKIPDIR = (".venv", ".nox", ".tox", "temp")
PYPROJECT = "pyproject.toml"
SETUP = "setup.py"

BumpTypes = {
    "M": "major",
    "m": "minor",
    "p": "patch",
    "b": "post",
    "c": "pre",
    "s": "skip",
}
BumpTypesPrompt = " | ".join([f"{k}-{v.title()}" for k, v in BumpTypes.items()])


class BumpType(Enum):
    MAJOR = "major"
    MINOR = "minor"
    PATCH = "patch"
    PRE = "pre"
    POST = "post"


def _get_bumped_version(current: VersionInfo, bump_type: BumpType) -> VersionInfo:
    if bump_type == BumpType.MAJOR:
        return current.next_version("major")

    if bump_type == BumpType.MINOR:
        return current.next_version("minor")

    if bump_type == BumpType.PATCH:
        return current.next_version("patch")

    if bump_type == BumpType.PRE:
        return current.next_version("prerelease")

    return current.bump_build(token="post")


def _skip(file: Path) -> bool:
    return any(skipdir in file.parts for skipdir in SKIPDIR)


def _find_packages():
    files = []
    for file in chain(CWD.glob(f"**/{PYPROJECT}"), CWD.glob(f"**/{SETUP}")):
        if _skip(file=file):
            continue
        files.append(file)
    return files


def _get_version_update(
    package: str,
    version: str,
    bump_type_selector: t.Optional[str] = None,
) -> t.Optional[str]:
    pre = None
    if "-" in version:
        version, pre = version.split("-", maxsplit=1)

    if "rc" in version:
        version, marker, number = version.partition("rc")
        pre = f"{marker}{number}"

    # TODO: Add support for post builds
    current_version = VersionInfo.parse(version=version)
    current_version._prerelease = pre
    bump_type_selector = t.cast(
        str,
        bump_type_selector
        or click.prompt(f"* Bump {package}@{version} -> {BumpTypesPrompt}"),
    )
    if bump_type_selector == "s":
        return None

    try:
        _bump_type = BumpType(BumpTypes[bump_type_selector])
    except (ValueError, KeyError) as e:
        raise click.ClickException(f"Invalid bump type: {e}")

    return str(_get_bumped_version(current=current_version, bump_type=_bump_type))


def _bump_setup_py(file: Path, bump_type: t.Optional[str] = None):
    # Load config
    content = file.read_text(encoding="utf-8")
    _, _, arguments = content.partition("setup(")

    package, version = None, None
    for line in arguments.split("\n"):
        if "=" not in line:
            continue

        arg, val = line.strip().split("=", maxsplit=1)
        val = val.replace('"', "").replace(",", "").strip()
        if arg == "name":
            package = val

        if arg == "version":
            version = val

    # Skip the main setup.py
    if package == "composio":
        return

    if package is None or version is None:
        raise ValueError(f"Package or version not found in {file}")

    # Bump version
    next_version = _get_version_update(
        package=package,
        version=version,
        bump_type_selector=bump_type,
    )
    if next_version is None:
        click.echo(f"* Skipping {package}")
        return

    click.echo(f"* Bumping {package} {version} -> {next_version}")
    content = content.replace(f'version="{version}"', f'version="{next_version}"')
    file.write_text(content, encoding="utf-8")


def _bump_pyproject_toml(file: Path, bump_type: t.Optional[str] = None):
    # Load config
    content = file.read_text(encoding="utf-8")
    config = tomli.loads(content)
    project = t.cast(dict, config["project"])

    # Extract meta
    package = project["name"]
    version = project["version"]

    # Bump version
    next_version = _get_version_update(
        package=package,
        version=version,
        bump_type_selector=bump_type,
    )
    if next_version is None:
        click.echo(f"* Skipping {package}")
        return

    click.echo(f"* Bumping {package} {version} -> {next_version}")
    content = content.replace(f'version = "{version}"', f'version = "{next_version}"')
    file.write_text(content, encoding="utf-8")


def _bump_package(file: Path, bump_type: t.Optional[str] = None):
    if file.name == PYPROJECT:
        return _bump_pyproject_toml(file=file, bump_type=bump_type)
    return _bump_setup_py(file=file, bump_type=bump_type)


@click.command("bump")
@click.option("--pre", "bump_type", flag_value="c")
@click.option("--post", "bump_type", flag_value="b")
@click.option("--patch", "bump_type", flag_value="p")
@click.option("--minor", "bump_type", flag_value="m")
@click.option("--major", "bump_type", flag_value="M")
def bump(bump_type: t.Optional[str]):
    packages = _find_packages()
    for package in packages:
        try:
            _bump_package(file=package, bump_type=bump_type)
        except Exception as e:
            raise e.__class__(f"Error bumping {package}")


if __name__ == "__main__":
    bump()



from __future__ import annotations

import functools
import json
import time
import traceback
import typing as t
import uuid
from concurrent.futures import ThreadPoolExecutor
from unittest import mock

import typing_extensions as te
from pysher import Pusher
from pysher.channel import Channel as PusherChannel
from pysher.connection import Connection as PusherConnection

from composio import exceptions
from composio.client import HttpClient
from composio.client.types import trigger_instance_upsert_response
from composio.core.models.base import Resource
from composio.core.models.internal import Internal
from composio.exceptions import ComposioSDKTimeoutError
from composio.utils.logging import WithLogger

PUSHER_AUTH_URL = "{base_url}/api/v3/internal/sdk/realtime/auth?source=python"

"""
export type _TriggerData = {
  appName: string;
  clientId: number;
  payload: Record<string, unknown>;
  originalPayload: Record<string, unknown>;
  metadata: {
    id: string;
    nanoId: string;
    triggerName: string;
    triggerData: string;
    triggerConfig: Record<string, unknown>;
    connection: {
      id: string;
      connectedAccountNanoId: string;
      integrationId: string;
      authConfigNanoId: string;
      clientUniqueUserId: string;
      status: string;
    };
  };
};
"""


class _ConnectionData(te.TypedDict):
    id: str
    status: t.Literal["ACTIVE", "INACTIVE"]
    integrationId: str
    authConfigNanoId: str
    clientUniqueUserId: str
    connectedAccountNanoId: str


class _TriggerMetadata(te.TypedDict):
    id: str
    nanoId: str
    triggerName: str
    triggerData: str
    triggerConfig: t.Dict
    connection: _ConnectionData


class _TriggerData(te.TypedDict):
    clientId: str
    appName: str
    payload: t.Dict
    metadata: _TriggerMetadata
    originalPayload: t.Dict


class TriggerConnectedAccountSchema(t.TypedDict):
    id: str
    uuid: str
    user_id: str
    auth_config_id: str
    auth_config_uuid: str
    status: t.Literal["ACTIVE", "INACTIVE"]


class TriggerMetadataSchema(t.TypedDict):
    id: str
    toolkit_slug: str
    trigger_slug: str
    trigger_data: t.Optional[str]
    trigger_config: t.Dict[str, t.Any]
    connected_account: TriggerConnectedAccountSchema


class TriggerEvent(t.TypedDict):
    id: str  # The ID of the trigger
    uuid: str  # UUID of the trigger
    user_id: str  # The ID of the user that triggered the event

    trigger_slug: str  # The slug of the trigger that triggered the event
    toolkit_slug: str  # The slug of the toolkit that triggered the event

    payload: t.Optional[t.Dict[str, t.Any]]  # The payload of the trigger
    metadata: TriggerMetadataSchema
    original_payload: t.Optional[
        t.Dict[str, t.Any]
    ]  # The original payload of the trigger


_ = {
    "appName": "github",
    "payload": {
        "author": "angrybayblade",
        "id": "a2334682759c4324e911d8f52f8fb6bdf1338d94",
        "message": "temp",
        "timestamp": "2025-06-19T11:28:01+05:30",
        "url": "https://github.com/angrybayblade/ph7/commit/a2334682759c4324e911d8f52f8fb6bdf1338d94",
    },
    "originalPayload": {
        "ref": "refs/heads/temp",
        "before": "e5eb72bc9f973b6cc82e0de9708660c5341befe5",
        "after": "a2334682759c4324e911d8f52f8fb6bdf1338d94",
        "repository": {
            "id": 752122629,
            "node_id": "R_kgDOLNR7BQ",
            "name": "ph7",
            "full_name": "angrybayblade/ph7",
            "private": False,
            "owner": {
                "name": "angrybayblade",
                "email": "35092918+angrybayblade@users.noreply.github.com",
                "login": "angrybayblade",
                "id": 35092918,
                "node_id": "MDQ6VXNlcjM1MDkyOTE4",
                "avatar_url": "https://avatars.githubusercontent.com/u/35092918?v=4",
                "gravatar_id": "",
                "url": "https://api.github.com/users/angrybayblade",
                "html_url": "https://github.com/angrybayblade",
                "followers_url": "https://api.github.com/users/angrybayblade/followers",
                "following_url": "https://api.github.com/users/angrybayblade/following{/other_user}",
                "gists_url": "https://api.github.com/users/angrybayblade/gists{/gist_id}",
                "starred_url": "https://api.github.com/users/angrybayblade/starred{/owner}{/repo}",
                "subscriptions_url": "https://api.github.com/users/angrybayblade/subscriptions",
                "organizations_url": "https://api.github.com/users/angrybayblade/orgs",
                "repos_url": "https://api.github.com/users/angrybayblade/repos",
                "events_url": "https://api.github.com/users/angrybayblade/events{/privacy}",
                "received_events_url": "https://api.github.com/users/angrybayblade/received_events",
                "type": "User",
                "user_view_type": "public",
                "site_admin": False,
            },
            "html_url": "https://github.com/angrybayblade/ph7",
            "description": "💧 Python native HTML rendering",
            "fork": False,
            "url": "https://api.github.com/repos/angrybayblade/ph7",
            "forks_url": "https://api.github.com/repos/angrybayblade/ph7/forks",
            "keys_url": "https://api.github.com/repos/angrybayblade/ph7/keys{/key_id}",
            "collaborators_url": "https://api.github.com/repos/angrybayblade/ph7/collaborators{/collaborator}",
            "teams_url": "https://api.github.com/repos/angrybayblade/ph7/teams",
            "hooks_url": "https://api.github.com/repos/angrybayblade/ph7/hooks",
            "issue_events_url": "https://api.github.com/repos/angrybayblade/ph7/issues/events{/number}",
            "events_url": "https://api.github.com/repos/angrybayblade/ph7/events",
            "assignees_url": "https://api.github.com/repos/angrybayblade/ph7/assignees{/user}",
            "branches_url": "https://api.github.com/repos/angrybayblade/ph7/branches{/branch}",
            "tags_url": "https://api.github.com/repos/angrybayblade/ph7/tags",
            "blobs_url": "https://api.github.com/repos/angrybayblade/ph7/git/blobs{/sha}",
            "git_tags_url": "https://api.github.com/repos/angrybayblade/ph7/git/tags{/sha}",
            "git_refs_url": "https://api.github.com/repos/angrybayblade/ph7/git/refs{/sha}",
            "trees_url": "https://api.github.com/repos/angrybayblade/ph7/git/trees{/sha}",
            "statuses_url": "https://api.github.com/repos/angrybayblade/ph7/statuses/{sha}",
            "languages_url": "https://api.github.com/repos/angrybayblade/ph7/languages",
            "stargazers_url": "https://api.github.com/repos/angrybayblade/ph7/stargazers",
            "contributors_url": "https://api.github.com/repos/angrybayblade/ph7/contributors",
            "subscribers_url": "https://api.github.com/repos/angrybayblade/ph7/subscribers",
            "subscription_url": "https://api.github.com/repos/angrybayblade/ph7/subscription",
            "commits_url": "https://api.github.com/repos/angrybayblade/ph7/commits{/sha}",
            "git_commits_url": "https://api.github.com/repos/angrybayblade/ph7/git/commits{/sha}",
            "comments_url": "https://api.github.com/repos/angrybayblade/ph7/comments{/number}",
            "issue_comment_url": "https://api.github.com/repos/angrybayblade/ph7/issues/comments{/number}",
            "contents_url": "https://api.github.com/repos/angrybayblade/ph7/contents/{+path}",
            "compare_url": "https://api.github.com/repos/angrybayblade/ph7/compare/{base}...{head}",
            "merges_url": "https://api.github.com/repos/angrybayblade/ph7/merges",
            "archive_url": "https://api.github.com/repos/angrybayblade/ph7/{archive_format}{/ref}",
            "downloads_url": "https://api.github.com/repos/angrybayblade/ph7/downloads",
            "issues_url": "https://api.github.com/repos/angrybayblade/ph7/issues{/number}",
            "pulls_url": "https://api.github.com/repos/angrybayblade/ph7/pulls{/number}",
            "milestones_url": "https://api.github.com/repos/angrybayblade/ph7/milestones{/number}",
            "notifications_url": "https://api.github.com/repos/angrybayblade/ph7/notifications{?since,all,participating}",
            "labels_url": "https://api.github.com/repos/angrybayblade/ph7/labels{/name}",
            "releases_url": "https://api.github.com/repos/angrybayblade/ph7/releases{/id}",
            "deployments_url": "https://api.github.com/repos/angrybayblade/ph7/deployments",
            "created_at": 1706936166,
            "updated_at": "2025-03-22T17:26:27Z",
            "pushed_at": 1750312684,
            "git_url": "git://github.com/angrybayblade/ph7.git",
            "ssh_url": "git@github.com:angrybayblade/ph7.git",
            "clone_url": "https://github.com/angrybayblade/ph7.git",
            "svn_url": "https://github.com/angrybayblade/ph7",
            "homepage": "https://angrybayblade.github.io/ph7/",
            "size": 3031,
            "stargazers_count": 7,
            "watchers_count": 7,
            "language": "Python",
            "has_issues": True,
            "has_projects": True,
            "has_downloads": True,
            "has_wiki": True,
            "has_pages": True,
            "has_discussions": False,
            "forks_count": 1,
            "mirror_url": None,
            "archived": False,
            "disabled": False,
            "open_issues_count": 0,
            "license": None,
            "allow_forking": True,
            "is_template": False,
            "web_commit_signoff_required": False,
            "topics": [
                "css",
                "django",
                "flask",
                "html",
                "js",
                "template-engine",
                "web",
            ],
            "visibility": "public",
            "forks": 1,
            "open_issues": 0,
            "watchers": 7,
            "default_branch": "main",
            "stargazers": 7,
            "master_branch": "main",
        },
        "pusher": {
            "name": "angrybayblade",
            "email": "35092918+angrybayblade@users.noreply.github.com",
        },
        "sender": {
            "login": "angrybayblade",
            "id": 35092918,
            "node_id": "MDQ6VXNlcjM1MDkyOTE4",
            "avatar_url": "https://avatars.githubusercontent.com/u/35092918?v=4",
            "gravatar_id": "",
            "url": "https://api.github.com/users/angrybayblade",
            "html_url": "https://github.com/angrybayblade",
            "followers_url": "https://api.github.com/users/angrybayblade/followers",
            "following_url": "https://api.github.com/users/angrybayblade/following{/other_user}",
            "gists_url": "https://api.github.com/users/angrybayblade/gists{/gist_id}",
            "starred_url": "https://api.github.com/users/angrybayblade/starred{/owner}{/repo}",
            "subscriptions_url": "https://api.github.com/users/angrybayblade/subscriptions",
            "organizations_url": "https://api.github.com/users/angrybayblade/orgs",
            "repos_url": "https://api.github.com/users/angrybayblade/repos",
            "events_url": "https://api.github.com/users/angrybayblade/events{/privacy}",
            "received_events_url": "https://api.github.com/users/angrybayblade/received_events",
            "type": "User",
            "user_view_type": "public",
            "site_admin": False,
        },
        "created": False,
        "deleted": False,
        "forced": False,
        "base_ref": None,
        "compare": "https://github.com/angrybayblade/ph7/compare/e5eb72bc9f97...a2334682759c",
        "commits": [
            {
                "id": "a2334682759c4324e911d8f52f8fb6bdf1338d94",
                "tree_id": "ed0ae8c902cb2a0e12d3c64aea8aab66f7b06054",
                "distinct": True,
                "message": "temp",
                "timestamp": "2025-06-19T11:28:01+05:30",
                "url": "https://github.com/angrybayblade/ph7/commit/a2334682759c4324e911d8f52f8fb6bdf1338d94",
                "author": {
                    "name": "angrybayblade",
                    "email": "vptl185@gmail.com",
                    "username": "angrybayblade",
                },
                "committer": {
                    "name": "angrybayblade",
                    "email": "vptl185@gmail.com",
                    "username": "angrybayblade",
                },
                "added": [],
                "removed": [],
                "modified": ["mkdocs.yml"],
            }
        ],
        "head_commit": {
            "id": "a2334682759c4324e911d8f52f8fb6bdf1338d94",
            "tree_id": "ed0ae8c902cb2a0e12d3c64aea8aab66f7b06054",
            "distinct": True,
            "message": "temp",
            "timestamp": "2025-06-19T11:28:01+05:30",
            "url": "https://github.com/angrybayblade/ph7/commit/a2334682759c4324e911d8f52f8fb6bdf1338d94",
            "author": {
                "name": "angrybayblade",
                "email": "vptl185@gmail.com",
                "username": "angrybayblade",
            },
            "committer": {
                "name": "angrybayblade",
                "email": "vptl185@gmail.com",
                "username": "angrybayblade",
            },
            "added": [],
            "removed": [],
            "modified": ["mkdocs.yml"],
        },
    },
    "metadata": {
        "id": "2507236e-4be2-4606-b338-fb24c76ce38d",
        "nanoId": "ti_VZEWUqF5fQ6P",
        "connectionId": "75271255-fe95-4d34-bee3-459a02645e38",
        "connectionNanoId": "ca_5KdA-e2C4ZMd",
        "triggerName": "GITHUB_COMMIT_EVENT",
        "triggerData": '{"event_type": "push", "github_hook_id": "552965247"}',
        "triggerConfig": {"repo": "ph7", "owner": "angrybayblade"},
        "connection": {
            "id": "75271255-fe95-4d34-bee3-459a02645e38",
            "connectedAccountNanoId": "ca_5KdA-e2C4ZMd",
            "integrationId": "500748ea-8547-4abb-9f9e-10dbcfdb81c2",
            "authConfigNanoId": "ac_ZxnpxqOo1nAP",
            "clientUniqueUserId": "default",
            "status": "ACTIVE",
        },
    },
}


class _ChunkedTriggerEventData(te.TypedDict):
    """Cunked trigger event data model."""

    id: str
    index: int
    chunk: str
    final: bool


class TriggerEventFilters(te.TypedDict):
    """Trigger event filterset."""

    trigger_slug: te.NotRequired[str]
    trigger_id: te.NotRequired[str]
    toolkit: te.NotRequired[str]
    user_id: te.NotRequired[str]
    auth_config_id: te.NotRequired[str]
    connected_account_id: te.NotRequired[str]


TriggerCallback = t.Callable[[TriggerEvent], None]


class TriggerSubscription(Resource):
    """Trigger subscription."""

    _pusher: Pusher
    _channel: PusherChannel
    _connection: PusherConnection
    _alive: bool

    def __init__(self, client: HttpClient) -> None:
        """Initialize subscription object."""
        super().__init__(client=client)
        self.client = client
        self._alive = False
        self._chunks: t.Dict[str, t.Dict[int, str]] = {}
        self._callbacks: t.List[t.Tuple[TriggerCallback, TriggerEventFilters]] = []

    def handle(
        self, **filters: te.Unpack[TriggerEventFilters]
    ) -> t.Callable[[TriggerCallback], TriggerCallback]:
        """Register a trigger callaback."""

        def _wrap(f: TriggerCallback) -> TriggerCallback:
            self.logger.debug(f"Registering callback `{f.__name__}`")
            self._callbacks.append((f, filters))
            return f

        return _wrap

    def _parse_payload(self, event: str) -> t.Optional[TriggerEvent]:
        """Parse event payload."""
        try:
            data = t.cast(_TriggerData, json.loads(event))
        except Exception as e:
            self.logger.warning(f"Error decoding payload: {e}")
            return None

        return t.cast(
            TriggerEvent,
            {
                "id": data["metadata"]["nanoId"],
                "uuid": data["metadata"]["id"],
                "user_id": data["metadata"]["connection"]["clientUniqueUserId"],
                "toolkit_slug": data["appName"],
                "trigger_slug": data["metadata"]["triggerName"],
                "metadata": {
                    "id": data["metadata"]["nanoId"],
                    "uuid": data["metadata"]["id"],
                    "toolkit_slug": data["appName"],
                    "trigger_slug": data["metadata"]["triggerName"],
                    "trigger_data": data["metadata"]["triggerData"],
                    "trigger_config": data["metadata"]["triggerConfig"],
                    "connected_account": {
                        "id": data["metadata"]["connection"]["connectedAccountNanoId"],
                        "uuid": data["metadata"]["connection"]["id"],
                        "auth_config_id": data["metadata"]["connection"][
                            "authConfigNanoId"
                        ],
                        "auth_config_uuid": data["metadata"]["connection"][
                            "integrationId"
                        ],
                        "user_id": data["metadata"]["connection"]["clientUniqueUserId"],
                        "status": data["metadata"]["connection"]["status"],
                    },
                },
                "payload": data["payload"],
                "original_payload": data["originalPayload"],
            },
        )

    def _handle_chunked_events(self, event: str) -> None:
        """Handle chunked events."""
        data = _ChunkedTriggerEventData(**json.loads(event))  # type: ignore
        if data["id"] not in self._chunks:
            self._chunks[data["id"]] = {}

        self._chunks[data["id"]][data["index"]] = data["chunk"]
        if data["final"]:
            _chunks = self._chunks.pop(data["id"])
            self._handle_event(event="".join([_chunks[idx] for idx in sorted(_chunks)]))

    def _filters_match(
        self,
        data: TriggerEvent,
        filters: TriggerEventFilters,
        callback: str,
    ) -> bool:
        """Check if filters match the event data."""
        checks = (
            ("trigger_slug", data["trigger_slug"]),
            ("trigger_id", data["metadata"]["id"]),
            ("toolkit", data["toolkit_slug"]),
            ("user_id", data["user_id"]),
            ("auth_config_id", data["metadata"]["connected_account"]["auth_config_id"]),
            ("connected_account_id", data["metadata"]["connected_account"]["id"]),
        )
        for name, check in checks:
            value = filters.get(name)
            if value is None or str(value).lower() == check.lower():
                continue

            self.logger.debug(
                f"Skipping `{callback}` since "
                f"`{name}` filter does not match the event metadata",
            )
            return False
        return True

    def _handle_callback(
        self,
        callback: TriggerCallback,
        data: TriggerEvent,
        filters: TriggerEventFilters,
    ) -> t.Any:
        """Handle callback."""
        if not self._filters_match(data, filters, callback.__name__):
            return

        try:
            callback(data)
        except Exception:
            self.logger.error(
                f"Error executing `{callback.__name__}` for "
                f"event `{data['metadata']['trigger_slug']}` "
                f"with error:\n {traceback.format_exc()}"
            )

    def _handle_event(self, event: str) -> None:
        """Filter events and call the callback function."""
        data = self._parse_payload(event=event)
        if data is None:
            self.logger.error(f"Error parsing trigger payload: {event}")
            return

        self.logger.debug(
            f"Received trigger event with trigger ID: {data['metadata']['id']} "
            f"and trigger name: {data['metadata']['trigger_slug']}"
        )
        awaitables: t.List = []
        with ThreadPoolExecutor() as executor:
            for callback, filters in self._callbacks:
                awaitables.append(
                    executor.submit(
                        self._handle_callback,
                        callback,
                        data,
                        filters,
                    )
                )
        _ = [future.result() for future in awaitables]

    def is_alive(self) -> bool:
        """Check if subscription is live."""
        return self._alive

    def has_errored(self) -> bool:
        """Check if the connection errored and disconnected."""
        return self._connection.socket is None or self._connection.socket.has_errored

    def set_alive(self) -> None:
        """Set `_alive` to True."""
        self._alive = True

    def wait_forever(self) -> None:
        """Wait infinitely."""
        while self.is_alive() and not self.has_errored():
            time.sleep(1)

    def stop(self) -> None:
        """Stop the trigger listener."""
        self._connection.disconnect()
        self._alive = False

    def restart(self) -> None:
        """Restart the subscription connection"""
        self._connection.disconnect()
        self._connection._connect()  # pylint: disable=protected-access


class _SubcriptionBuilder(WithLogger):
    """Pusher client for Composio SDK."""

    def __init__(self, client: HttpClient) -> None:
        """Initialize pusher client."""
        super().__init__()
        self._client = client
        self.api_key = self._client.api_key
        self.base_url = self._client.base_url
        self.internal = Internal(client=self._client)
        self.subscription = TriggerSubscription(client=self._client)

    def _get_connection_handler(
        self,
        project_id: str,
        pusher: Pusher,
        subscription: TriggerSubscription,
    ) -> t.Callable[[str], None]:
        def _connection_handler(_: str) -> None:
            channel = t.cast(
                PusherChannel,
                pusher.subscribe(
                    channel_name=f"private-{project_id}_triggers",
                ),
            )
            channel.bind(
                event_name="trigger_to_client",
                callback=subscription._handle_event,
            )
            channel.bind(
                event_name="chunked-trigger_to_client",
                callback=subscription._handle_chunked_events,
            )
            subscription.set_alive()
            subscription._channel = channel  # pylint: disable=protected-access
            subscription._connection = (  # pylint: disable=protected-access
                channel.connection
            )

        return _connection_handler

    def _get_pusher_instance(self, key: str, cluster: str) -> Pusher:
        """Get a pusher instance."""
        return Pusher(
            key=key,
            cluster=cluster,
            auth_endpoint=PUSHER_AUTH_URL.format(base_url=self._client.base_url),
            auth_endpoint_headers={
                "x-api-key": self._client.api_key,
                "x-request-id": str(uuid.uuid4()),
            },
            auto_sub=True,
        )

    def connect(self, timeout: float = 15.0) -> TriggerSubscription:
        """Connect to Pusher channel for given client ID."""
        self.logger.debug("Creating trigger subscription")
        project_info = self.internal.get_sdk_realtime_credentials()
        pusher = self._get_pusher_instance(
            key=project_info.pusher_key,
            cluster=project_info.pusher_cluster,
        )

        # Patch pusher logger
        pusher.connection.logger = mock.MagicMock()  # type: ignore
        pusher.connection.bind(
            "pusher:connection_established",
            self._get_connection_handler(
                project_id=project_info.project_id,
                pusher=pusher,
                subscription=self.subscription,
            ),
        )
        pusher.connect()

        # Wait for connection to get established
        deadline = time.time() + timeout
        while time.time() < deadline:
            if not self.subscription.is_alive():
                time.sleep(0.5)
                continue

            self.subscription._pusher = pusher  # pylint: disable=protected-access
            return self.subscription
        raise ComposioSDKTimeoutError(
            "Timed out while waiting for trigger listener to be established"
        )


class Triggers(Resource):
    """Triggers (instance) class"""

    enable: t.Callable
    """Enables a trigger given its id"""

    disable: t.Callable
    """Disables a trigger given its id"""

    def __init__(self, client: HttpClient):
        """
        Initialize the triggers resource.

        :param client: The client to use for the triggers resource.
        """
        self._client = client
        self.list_enum = self._client.triggers_types.retrieve_enum
        self.get_type = self._client.triggers_types.retrieve
        self.list = self._client.triggers_types.list
        self.delete = self._client.trigger_instances.manage.delete
        self.enable = functools.partial(
            self._client.trigger_instances.manage.update,
            status="enable",
        )
        self.disable = functools.partial(
            self._client.trigger_instances.manage.update,
            status="disable",
        )

    def list_active(
        self,
        trigger_ids: t.Optional[list[str]] = None,
        trigger_names: t.Optional[list[str]] = None,
        auth_config_ids: t.Optional[list[str]] = None,
        connected_account_ids: t.Optional[list[str]] = None,
        show_disabled: t.Optional[bool] = None,
        limit: t.Optional[int] = None,
        page: t.Optional[int] = None,
    ):
        """
        List all active triggers

        :param trigger_ids: List of trigger IDs to filter by
        :param trigger_names: List of trigger names to filter by
        :param auth_config_ids: List of auth config IDs to filter by
        :param connected_account_ids: List of connected account IDs to filter by
        :param show_disabled: Whether to show disabled triggers
        :param limit: Limit the number of triggers to return
        :param page: Page number to return
        :return: List of active triggers
        """
        return self._client.trigger_instances.list_active(
            query_trigger_ids_1=trigger_ids,
            query_trigger_names_1=trigger_names,
            query_auth_config_ids_1=auth_config_ids,
            query_connected_account_ids_1=connected_account_ids,
            query_show_disabled_1=show_disabled,
            limit=limit or self._client.not_given,
            page=page or self._client.not_given,
        )

    @t.overload
    def create(
        self,
        slug: str,
        *,
        connected_account_id: str,
        trigger_config: t.Optional[t.Dict[str, t.Any]] = None,
    ) -> trigger_instance_upsert_response.TriggerInstanceUpsertResponse: ...

    @t.overload
    def create(
        self,
        slug: str,
        *,
        user_id: str,
        trigger_config: t.Optional[t.Dict[str, t.Any]] = None,
    ) -> trigger_instance_upsert_response.TriggerInstanceUpsertResponse: ...

    def create(
        self,
        slug: str,
        *,
        user_id: t.Optional[str] = None,
        connected_account_id: t.Optional[str] = None,
        trigger_config: t.Optional[t.Dict[str, t.Any]] = None,
    ) -> trigger_instance_upsert_response.TriggerInstanceUpsertResponse:
        """
        Create a trigger instance

        :param slug: The slug of the trigger
        :param connected_account_id: The ID of the connected account
        :param trigger_config: The configuration of the trigger
        :return: The trigger instance
        """
        if user_id is not None:
            connected_account_id = self._get_connected_account_for_user(
                trigger=slug,
                user_id=user_id,
            )

        if connected_account_id is None:
            raise exceptions.InvalidParams(
                "please provide valid `connected_account` or `user_id`"
            )

        return self._client.trigger_instances.upsert(
            slug=slug,
            connected_account_id=connected_account_id,
            body_trigger_config_1=trigger_config or self._client.not_given,
        )

    def _get_connected_account_for_user(self, trigger: str, user_id: str) -> str:
        toolkit = self.get_type(slug=trigger).toolkit.slug
        connected_accounts = self._client.connected_accounts.list(
            toolkit_slugs=[toolkit]
        )
        if len(connected_accounts.items) == 0:
            raise exceptions.NoItemsFound(
                f"No connected accounts found for {trigger} and {user_id}"
            )

        account, *_ = sorted(
            connected_accounts.items,
            key=lambda x: x.created_at,
            reverse=True,
        )
        return account.id

    def subscribe(self, timeout: float = 15.0) -> TriggerSubscription:
        """
        Subscribe to a trigger and receive trigger events.

        :param timeout: The timeout to wait for the subscription to be established.
        :return: The trigger subscription handler.
        """
        return _SubcriptionBuilder(client=self._client).connect(timeout=timeout)



from composio import Composio

composio = Composio()

# Get all tools by toolkit
tools = composio.tools.get(user_id="default", toolkits=["GITHUB"])

# Get all tools by search
tools = composio.tools.get(user_id="default", search="user")

# Get all tools by toolkit and search
tools = composio.tools.get(user_id="default", toolkits=["GITHUB"], search="star")

# execute tool
response = composio.tools.execute(
    user_id="default",
    slug="HACKERNEWS_GET_USER",
    arguments={"username": "pg"},
)
print(response)

# execute proxy call (github)
proxy_response = composio.tools.proxy(
    endpoint="/repos/composiohq/composio/issues/1",
    method="GET",
    connected_account_id="ac_1234",  # use connected account for github
    parameters=[
        {
            "name": "Accept",
            "value": "application/vnd.github.v3+json",
            "type": "header",
        },
    ],
)
print(proxy_response)



"""
This module is a light wrapper around the auto-generated composio client.
"""

import contextvars
import typing as t
from uuid import uuid4

import typing_extensions as te
from composio_client import (
    DEFAULT_MAX_RETRIES,
    NOT_GIVEN,
    APIError,
    NotGiven,
    _base_client,
)
from composio_client import Composio as BaseComposio
from httpx import URL, Client, Request, Timeout

from composio.utils.logging import WithLogger

ComposioAPIError = APIError
APIEnvironment = te.Literal["production", "staging", "local"]


class RequestContext(te.TypedDict):
    id: te.NotRequired[t.Optional[str]]
    provider: str


# TODO: Rename `Composio` to `HttpClient` in stainless generator
class HttpClient(BaseComposio, WithLogger):
    """
    Wrapper around the auto-generated composio client.
    """

    request_ctx: contextvars.ContextVar[RequestContext]
    not_given = NOT_GIVEN

    def __init__(
        self,
        *,
        provider: str,
        api_key: t.Optional[str] = None,
        environment: te.Union[NotGiven, APIEnvironment] = "production",
        base_url: t.Optional[t.Union[str, URL, NotGiven]] = NOT_GIVEN,
        timeout: t.Optional[t.Union[float, Timeout, NotGiven]] = NOT_GIVEN,
        max_retries: int = DEFAULT_MAX_RETRIES,
        default_headers: t.Optional[t.Mapping[str, str]] = None,
        default_query: t.Optional[t.Mapping[str, object]] = None,
        http_client: t.Optional[Client] = None,
        _strict_response_validation: bool = False,
    ) -> None:
        """
        Initialize the client.

        :param provider: The provider to use for the client.
        :param api_key: The API key to use for the client.
        :param environment: The environment to use for the client.
        :param base_url: The base URL to use for the client.
        :param timeout: The timeout to use for the client.
        :param max_retries: The maximum number of retries to use for the client.
        :param default_headers: The default headers to use for the client.
        :param default_query: The default query parameters to use for the client.
        :param http_client: The HTTP client to use for the client.
        """
        WithLogger.__init__(self)
        BaseComposio.__init__(
            self,
            api_key=api_key,
            environment=environment,
            base_url=base_url,
            timeout=timeout,
            max_retries=max_retries,
            default_headers=default_headers,
            default_query=default_query,
            http_client=http_client,
            _strict_response_validation=_strict_response_validation,
        )
        # TOFIX: Verbosity wrapper impl
        _base_client.log = self._logger  # type: ignore
        self.provider = provider
        self.request_ctx = contextvars.ContextVar[RequestContext](
            "request_ctx",
            default={
                "id": None,
                "provider": provider,
            },
        )

    def _prepare_request(self, request: Request) -> None:
        """
        Request interceptor to inject request id and provider.
        """
        ctx = self.request_ctx.get()
        request.headers["x-request-id"] = ctx.get("id") or uuid4().hex
        request.headers["x-framework-provider"] = ctx["provider"]



from composio import Composio
from composio.types import auth_scheme

composio = Composio()

# List all connected accounts
connected_accounts = composio.connected_accounts.list()
print(connected_accounts)

# Create a new connected account (OAuth)
connection_request = composio.connected_accounts.initiate(
    user_id="1234567890",
    auth_config_id="1234567890",
)
print(connection_request)

# Wait for the connection to be established (OAuth)
connected_account = connection_request.wait_for_connection()
print(connected_account)

# Create a new connected account (API Key)
connection_request = composio.connected_accounts.initiate(
    user_id="1234567890",
    auth_config_id="1234567890",
    config=auth_scheme.api_key(
        options={
            "api_key": "1234567890",
        },
    ),
)
print(connection_request)

# When creating a connected account, you can check for required fields
required_fields = composio.toolkits.get_connected_account_initiation_fields(
    toolkit="NOTION",
    auth_scheme="API_KEY",
)
print(required_fields)

# Retrieve a specific connected account
connected_account_retrieved = composio.connected_accounts.get(connected_account.id)
print(connected_account_retrieved)

# Disable a connected account
composio.connected_accounts.disable(connected_account.id)
print("Connected account disabled")

# Enable a connected account
composio.connected_accounts.enable(connected_account.id)
print("Connected account enabled")

# Delete a connected account
composio.connected_accounts.delete(connected_account.id)
print("Connected account deleted")



"""OpenAPI helpers."""

import inspect
import typing as t

from composio.exceptions import InvalidSchemaError

OPENAPI_TO_PYTHON = {
    "null": None,
    "number": float,
    "integer": int,
    "boolean": bool,
    "string": str,
}


# pylint: disable=unused-argument
def _handle_object_type(schema: t.Dict) -> t.Type:
    # Nested objects are not supported ATM
    return t.Dict[str, t.Any]


def _handle_array_type(schema: t.Dict) -> t.Any:
    # This discards the nested objects
    items_type = schema.get("items", {}).get("type")
    if items_type is None:
        return t.List[t.Any]

    SubT = _type_to_parameter(schema=schema.get("items", {}))
    return t.List[SubT]  # type: ignore


def _handle_enum_type(schema: t.Dict) -> t.Any:
    return t.Literal[tuple(schema["enum"])]


def _type_to_parameter(schema: t.Dict[str, t.Any]) -> t.Any:
    if "enum" in schema:
        return _handle_enum_type(schema=schema)

    p_type = schema["type"]
    if p_type in OPENAPI_TO_PYTHON:
        return OPENAPI_TO_PYTHON[p_type]

    if p_type == "object":
        return _handle_object_type(schema=schema)

    if p_type == "array":
        return _handle_array_type(schema=schema)

    raise InvalidSchemaError(f"Invalid property type {p_type}: {schema!r}")


def _handle_composite_type(schemas: t.List[t.Dict]) -> t.Any:
    return t.Union[tuple(map(_type_to_parameter, schemas))]


def _one_of_to_parameter(schema: t.Dict[str, t.Any]) -> t.Any:
    return _handle_composite_type(schemas=schema["oneOf"])


def _any_of_to_parameter(schema: t.Dict[str, t.Any]) -> t.Any:
    return _handle_composite_type(schemas=schema["anyOf"])


def _all_of_to_parameter(schema: t.Dict[str, t.Any]) -> t.Type:
    composite = {}
    for subschema in schema["allOf"]:
        composite.update(subschema)
    return _type_to_parameter(schema=composite)


def function_signature_from_jsonschema(
    schema: t.Dict[str, t.Any],
    skip_default: bool = False,
) -> t.List[inspect.Parameter]:
    """Convert json schema to a list of parameters (`inspect.Parameter`)."""
    parameters = []
    required = set(schema.get("required", []))
    for p_name, p_schema in schema.get("properties", {}).items():
        if "oneOf" in p_schema:
            p_type = _one_of_to_parameter(schema=p_schema)
        elif "anyOf" in p_schema:
            p_type = _any_of_to_parameter(schema=p_schema)
        elif "allOf" in p_schema:
            p_type = _all_of_to_parameter(schema=p_schema)
        elif "type" in p_schema:
            p_type = _type_to_parameter(schema=p_schema)
        else:
            raise InvalidSchemaError(f"Invalid property object {p_name}: {p_schema!r}")

        p_val = p_schema.get("default", None)
        if p_name in required or p_schema.get("required", False) or skip_default:
            p_val = inspect.Parameter.empty

        parameters.append(
            inspect.Parameter(
                name=p_name,
                annotation=p_type,
                default=p_val,
                kind=inspect.Parameter.POSITIONAL_OR_KEYWORD,
            )
        )

    return parameters



class DeprecationError(Exception):
    """Raised when deprecating some older functions. This is strictly for use
    while developing and will be removed from the codebase later."""

    pass


def deprecate(reason: str = "This function is deprecated"):
    """Deprecation decorator. Provide `reason` to show why you're deprecating something.
    NOTE: Decorating something with this will ensure that the function _will not run._
    """
    import functools

    def decorator(func):
        @functools.wraps(func)
        def wrapper(*args, **kwargs):
            raise DeprecationError(f"{func.__name__} is deprecatiod: `{reason}`")

        return wrapper

    return decorator



"""
Langchain demo.
"""

from composio import Composio
from composio_langchain import LangchainProvider

from langchain import hub  # type: ignore
from langchain.agents import AgentExecutor, create_openai_functions_agent
from langchain_openai import ChatOpenAI


# Pull relevant agent model.
prompt = hub.pull("hwchase17/openai-functions-agent")

# Initialize tools.
openai_client = ChatOpenAI(model="gpt-4-turbo")


def main():
    composio = Composio(provider=LangchainProvider())

    # Get All the tools
    tools = composio.tools.get(user_id="default", toolkits=["GITHUB"])

    # Define task
    task = "Star a repo composiohq/composio on GitHub"

    # Define agent
    agent = create_openai_functions_agent(openai_client, tools, prompt)
    agent_executor = AgentExecutor(agent=agent, tools=tools, verbose=True)

    # Execute using agent_executor
    agent_executor.invoke({"input": task})


if __name__ == "__main__":
    main()



"""Custom Tools module"""

import functools
import inspect
import typing as t

from pydantic import BaseModel

from composio.client import HttpClient, NotGiven
from composio.client.types import (
    Tool,
    tool_list_response,
    tool_proxy_params,
    tool_proxy_response,
)
from composio.exceptions import (
    NotFoundError,
)


class ExecuteRequestFn(t.Protocol):
    def __call__(
        self,
        endpoint: str,
        method: t.Literal["GET", "POST", "PUT", "DELETE", "PATCH"],
        body: t.Dict | NotGiven = HttpClient.not_given,
        connected_account_id: str | NotGiven = HttpClient.not_given,
        parameters: t.Iterable[tool_proxy_params.Parameter]
        | NotGiven = HttpClient.not_given,
    ) -> tool_proxy_response.ToolProxyResponse: ...


class CustomToolProtocol(t.Protocol):
    def __call__(self, request: t.Any) -> t.Any: ...


class CustomToolWithProxyProtocol(t.Protocol):
    def __call__(
        self,
        request: t.Any,
        execute_request: ExecuteRequestFn,
        auth_credentials: t.Dict,
    ) -> t.Any: ...


CustomToolFn = t.Union[
    CustomToolProtocol,
    CustomToolWithProxyProtocol,
]


class CustomTool:
    request_model: t.Type[BaseModel]

    def __init__(
        self,
        f: t.Callable[..., t.Any],
        client: HttpClient,
        toolkit: t.Optional[str] = None,
    ):
        """
        Initialize the custom tool.

        :param f: The function to wrap.
        :param client: The client to use for the custom tool.
        :param toolkit: The toolkit to use for the custom tool.
        """
        self.f = t.cast(CustomToolFn, f)
        self.name, self.slug = f.__name__, f.__name__.upper()
        self.description = t.cast(str, f.__doc__)
        self.toolkit = toolkit
        self.client = client
        self.info = self.__parse_info()
        if self.toolkit is not None:
            self.name = f"{self.toolkit.lower()}_{self.name}"
            self.slug = f"{self.toolkit.upper()}_{self.slug}"

    def __parse_info(self) -> Tool:
        """Parse the parameters of the custom tool."""
        if self.description is None:
            raise ValueError(f"Description is required on tool {self.name}")

        self.request_model = inspect.signature(self.f).parameters["request"].annotation
        return Tool(
            name=self.name,
            description=self.description,
            input_parameters={},
            output_parameters={},
            available_versions=[],
            version="1.0.0",
            scopes=[],
            slug=self.slug,
            toolkit=tool_list_response.ItemToolkit(
                logo="",
                name=self.toolkit or "custom",
                slug=(self.toolkit or "custom").upper(),
            ),
            deprecated=tool_list_response.ItemDeprecated(
                available_versions=[],
                displayName=self.name,
                version="1.0.0",
                toolkit=tool_list_response.ItemDeprecatedToolkit(logo=str()),
                is_deprecated=False,
            ),
            no_auth=False,
            tags=[],
        )

    def __get_auth_credentials(self, user_id: str) -> dict:
        """Get the auth config for the custom tool."""
        if self.toolkit is None:
            raise ValueError("Toolkit is required for custom tools")

        connected_accounts = self.client.connected_accounts.list(
            toolkit_slugs=[self.toolkit],
            user_ids=[user_id],
        )

        if len(connected_accounts.items) == 0:
            raise ValueError(
                f"No connected accounts found for toolkit {self.toolkit} and user {user_id}"
            )

        account, *_ = sorted(
            connected_accounts.items,
            key=lambda x: x.created_at,
            reverse=True,
        )
        return account.state.val.model_dump()

    def __call__(self, **kwargs: t.Any) -> t.Any:
        """Call the custom tool."""
        user_id = kwargs.pop("user_id", "default")
        request = self.request_model.model_validate(kwargs)
        if self.toolkit is None:
            return t.cast(CustomToolProtocol, self.f)(request=request)

        return t.cast(CustomToolWithProxyProtocol, self.f)(
            request=request,
            execute_request=t.cast(ExecuteRequestFn, self.client.tools.proxy),
            auth_credentials=self.__get_auth_credentials(user_id),
        )


class CustomTools:
    """CustomTools class

    Used to manage custom tools created by users."""

    def __init__(self, client: HttpClient):
        self.client = client
        self.custom_tools_registry: dict[str, CustomTool] = {}

    def __getitem__(self, slug: str) -> CustomTool:
        """Get a custom tool by its slug."""
        return self.custom_tools_registry[slug]

    def get(self, slug: str) -> t.Optional[CustomTool]:
        """Get a custom tool by its slug."""
        try:
            return self.custom_tools_registry[slug]
        except KeyError:
            return None

    @t.overload
    def register(self, f: t.Optional[CustomToolProtocol] = None) -> t.Callable: ...

    @t.overload
    def register(
        self, *, toolkit: t.Optional[str] = None
    ) -> t.Callable[[CustomToolWithProxyProtocol], CustomTool]: ...

    def register(
        self,
        f: t.Optional[CustomToolProtocol] = None,
        *,
        toolkit: t.Optional[str] = None,
    ) -> t.Union[
        CustomTool,
        t.Callable[[CustomToolProtocol], CustomTool],
        t.Callable[[CustomToolWithProxyProtocol], CustomTool],
    ]:
        """Register a custom tool."""
        if f is not None:
            return self._wrap_tool(f, toolkit)
        return functools.partial(self._wrap_tool, toolkit=toolkit)

    def _wrap_tool(
        self,
        f: CustomToolFn,
        toolkit: t.Optional[str] = None,
    ) -> CustomTool:
        """Wrap a tool function."""
        tool = CustomTool(f=f, client=self.client, toolkit=toolkit)
        self.custom_tools_registry[tool.slug] = tool
        return tool

    def execute(self, slug: str, request: t.Dict) -> t.Dict:
        """Execute a custom tool."""
        custom_tool = self.get(slug)
        if custom_tool is None:
            raise NotFoundError(f"Custom tool with slug {slug} not found")
        return custom_tool(**request)



import pytest


@pytest.mark.openai
def test_openai_get_tools():
    """we can get tools with the OpenAIProvider"""
    from composio_openai import OpenAIProvider

    from composio.core import Composio

    composio = Composio(provider=OpenAIProvider())
    # Get GitHub tools that are pre-configured
    tools = composio.tools.get(user_id="default", filters={"toolkits": ["GITHUB"]})
    assert isinstance(tools, list)
    assert "function" in tools[0]
    assert "type" in tools[0]
    assert tools[0].get("type") == "function"


@pytest.mark.openai
def test_openai_get_multiple_tools():
    """We can get multiple tools with the OpenAIProvider"""
    from composio_openai import OpenAIProvider

    from composio.core import Composio

    composio_openai_provider = OpenAIProvider()
    composio = Composio(provider=composio_openai_provider)
    tools = composio.tools.get(
        user_id="default", filters={"category": "search", "limit": 10}
    )
    print(tools)


@pytest.mark.openai
def test_openai_handle_tool_calls():
    """Handles tool calls after getting tools"""
    from composio_openai import OpenAIProvider
    from openai import OpenAI

    from composio.core import Composio

    # Initialize tools.
    openai_client = OpenAI()
    composio = Composio(provider=OpenAIProvider())

    # Get GitHub tools that are pre-configured
    tools = composio.tools.get(user_id="default", filters={"toolkits": ["HACKERNEWS"]})
    # Get response from the LLM
    response = openai_client.chat.completions.create(
        model="gpt-4-turbo-preview",
        tools=tools,
        messages=[
            {"role": "system", "content": "You are a helpful assistant."},
            {
                "role": "user",
                "content": "Get the user details for `dang` on Hackernews.",
            },
        ],
    )
    # Execute the function calls.
    result = composio.provider.handle_tool_calls(response=response, user_id="default")
    assert result[0].data["response_data"]["username"] == "dang"
    print(result)



"""
OpenAI demo.
"""

from openai import OpenAI
from composio import Composio
from composio_openai import OpenAIProvider


# Initialize tools.
openai_client = OpenAI()
composio = Composio(provider=OpenAIProvider())

# Define task.
task = "Star a repo composiohq/composio on GitHub"

# Get GitHub tools that are pre-configured
tools = composio.tools.get(user_id="default", toolkits=["GITHUB"])

# Get response from the LLM
response = openai_client.chat.completions.create(
    model="gpt-4o-mini",
    tools=tools,
    messages=[
        {"role": "system", "content": "You are a helpful assistant."},
        {"role": "user", "content": task},
    ],
)
print(response)

# Execute the function calls.
result = composio.provider.handle_tool_calls(response=response, user_id="default")
print(result)




from __future__ import annotations

import functools
import typing as t

import typing_extensions as te

from composio.client import HttpClient
from composio.client.types import (
    Tool,
    tool_execute_params,
    tool_proxy_params,
    tool_proxy_response,
)
from composio.core.models._files import FileHelper
from composio.core.models.base import Resource
from composio.core.models.custom_tools import CustomTools
from composio.core.provider import (
    TProvider,
)
from composio.core.provider.agentic import AgenticProvider, AgenticProviderExecuteFn
from composio.core.provider.none_agentic import (
    NonAgenticProvider,
    NoneAgenticProviderExecuteFn,
)
from composio.exceptions import InvalidParams, NotFoundError

from ._modifiers import (
    Modifiers,
    ToolExecuteParams,
    after_execute,
    apply_modifier_by_type,
    before_execute,
    schema_modifier,
)


class ToolExecutionResponse(te.TypedDict):
    data: t.Dict
    error: t.Optional[str]
    successful: bool


class Tools(Resource, t.Generic[TProvider]):
    """
    Tools class definition

    This class is used to manage tools in the Composio SDK.
    It provides methods to list, get, and execute tools.
    """

    provider: TProvider

    def __init__(self, client: HttpClient, provider: TProvider):
        """
        Initialize the tools resource.

        :param client: The client to use for the tools resource.
        :param provider: The provider to use for the tools resource.
        """
        self._client = client
        self._custom_tools = CustomTools(client)
        self._tool_schemas: t.Dict[str, Tool] = {}
        self._file_helper = FileHelper(client=self._client)

        self.custom_tool = self._custom_tools.register
        self.provider = provider

    def _filter_custom_tools(
        self, tools: t.List[str]
    ) -> t.Tuple[t.List[str], t.List[Tool]]:
        """Filter out custom tools from the list of tools."""
        _tools = []
        _custom_tools = []
        for tool in tools:
            try:
                _custom_tools.append(t.cast(Tool, self._custom_tools[tool]))
            except KeyError:
                _tools.append(tool)
        return _tools, _custom_tools

    def get_raw_composio_tool_by_slug(self, slug: str) -> Tool:
        """
        Returns schema for the given tool slug.
        """
        try:
            return t.cast(Tool, self._custom_tools[slug])
        except KeyError:
            return t.cast(Tool, self._client.tools.retrieve(tool_slug=slug))

    def get_raw_composio_tools(
        self,
        tools: t.Optional[list[str]] = None,
        search: t.Optional[str] = None,
        toolkits: t.Optional[list[str]] = None,
        scopes: t.Optional[t.List[str]] = None,
        limit: t.Optional[int] = None,
    ) -> list[Tool]:
        """
        Get a list of tool schemas based on the provided filters.
        """
        if tools is None and search is None and toolkits is None:
            raise InvalidParams(
                "Either `tools`, `search`, or `toolkits` must be provided"
            )

        tools_list = []
        if tools is not None:
            tools, custom_tools = self._filter_custom_tools(tools=tools)
            tools_list.extend(custom_tools)
            tools_list.extend(self._client.tools.list(tool_slugs=",".join(tools)).items)

        # Search tools by toolkit slugs and search term
        if toolkits is not None or search is not None:
            tools_list.extend(
                self._client.tools.list(
                    toolkit_slug=(
                        ",".join(toolkits) if toolkits else self._client.not_given
                    ),
                    search=search if search else self._client.not_given,
                    scopes=scopes,
                    limit=str(limit) if limit is not None else self._client.not_given,
                ).items
            )
        return tools_list

    def _get(
        self,
        user_id: str,
        tools: t.Optional[list[str]] = None,
        search: t.Optional[str] = None,
        toolkits: t.Optional[list[str]] = None,
        scopes: t.Optional[t.List[str]] = None,
        modifiers: t.Optional[Modifiers] = None,
        limit: t.Optional[int] = None,
    ):
        """Get a list of tools based on the provided filters."""
        tools_list = self.get_raw_composio_tools(
            tools=tools,
            search=search,
            toolkits=toolkits,
            scopes=scopes,
            limit=limit,
        )
        if modifiers is not None:
            tools_list = [
                apply_modifier_by_type(
                    modifiers=modifiers,
                    toolkit=tool.toolkit.slug,
                    tool=tool.slug,
                    type="schema",
                    schema=tool,
                )
                for tool in tools_list
            ]

        self._tool_schemas.update({tool.slug: tool for tool in tools_list})
        for tool in tools_list:
            tool.input_parameters = self._file_helper.process_schema_recursively(
                schema=tool.input_parameters,
            )

        if issubclass(type(self.provider), NonAgenticProvider):
            t.cast(NonAgenticProvider, self.provider).set_execute_tool_fn(
                t.cast(
                    NoneAgenticProviderExecuteFn,
                    self.execute,
                )
            )
            return t.cast(NonAgenticProvider, self.provider).wrap_tools(
                tools=tools_list
            )

        return t.cast(AgenticProvider, self.provider).wrap_tools(
            tools=tools_list,
            execute_tool=self._wrap_execute_tool(
                user_id=user_id,
                modifiers=modifiers,
            ),
        )

    @t.overload
    def get(
        self,
        user_id: str,
        *,
        slug: str,
        modifiers: t.Optional[Modifiers] = None,
    ):
        """Get tool by slug"""

    @t.overload
    def get(
        self,
        user_id: str,
        *,
        tools: list[str],
        modifiers: t.Optional[Modifiers] = None,
    ):
        """Get tools by tool slugs"""

    @t.overload
    def get(
        self,
        user_id: str,
        *,
        toolkits: list[str],
        scopes: t.Optional[t.List[str]] = None,
        limit: t.Optional[int] = None,
        modifiers: t.Optional[Modifiers] = None,
    ):
        """Get tools by toolkit slugs (Only important tools are returned)"""

    @t.overload
    def get(
        self,
        user_id: str,
        *,
        search: str,
        modifiers: t.Optional[Modifiers] = None,
    ):
        """Search tool by search term"""

    @t.overload
    def get(
        self,
        user_id: str,
        *,
        toolkits: list[str],
        search: t.Optional[str] = None,
        limit: t.Optional[int] = None,
        modifiers: t.Optional[Modifiers] = None,
    ):
        """Get tool by search term and/or toolkit slugs and search term"""

    def get(
        self,
        user_id: str,
        *,
        slug: t.Optional[str] = None,
        tools: t.Optional[list[str]] = None,
        search: t.Optional[str] = None,
        toolkits: t.Optional[list[str]] = None,
        scopes: t.Optional[t.List[str]] = None,
        modifiers: t.Optional[Modifiers] = None,
        limit: t.Optional[int] = None,
    ):
        """Get a tool or list of tools based on the provided arguments."""
        if slug is not None:
            return self._get(user_id=user_id, tools=[slug], modifiers=modifiers)
        return self._get(
            user_id=user_id,
            tools=tools,
            search=search,
            toolkits=toolkits,
            scopes=scopes,
            modifiers=modifiers,
            limit=limit,
        )

    def _wrap_execute_tool(
        self,
        modifiers: t.Optional[Modifiers] = None,
        user_id: t.Optional[str] = None,
    ) -> AgenticProviderExecuteFn:
        """Wrap the execute tool function"""
        return t.cast(
            AgenticProviderExecuteFn,
            functools.partial(
                self.execute,
                modifiers=modifiers,
                user_id=user_id,
            ),
        )

    def _execute_custom_tool(
        self,
        slug: str,
        arguments: t.Dict,
    ) -> ToolExecutionResponse:
        """Execute a custom tool"""
        # TODO: Better error handling, pydantic validation eg...
        try:
            return {
                "data": self._custom_tools.execute(slug=slug, request=arguments),
                "error": None,
                "successful": True,
            }
        except NotFoundError:
            return {
                "data": {},
                "error": f"Tool with slug {slug} not found",
                "successful": False,
            }

    def _execute_tool(
        self,
        slug: str,
        arguments: t.Dict,
        connected_account_id: t.Optional[str] = None,
        custom_auth_params: t.Optional[tool_execute_params.CustomAuthParams] = None,
        user_id: t.Optional[str] = None,
        text: t.Optional[str] = None,
        version: t.Optional[str] = None,
    ) -> ToolExecutionResponse:
        """Execute a tool"""
        return t.cast(
            ToolExecutionResponse,
            self._client.tools.execute(
                tool_slug=slug,
                arguments=arguments,
                connected_account_id=connected_account_id or self._client.not_given,
                custom_auth_params=custom_auth_params or self._client.not_given,
                user_id=user_id or self._client.not_given,
                text=text or self._client.not_given,
                version=version or self._client.not_given,
            ).model_dump(
                exclude={
                    "log_id",
                    "session_info",
                }
            ),
        )

    def execute(
        self,
        slug: str,
        arguments: t.Dict,
        *,
        connected_account_id: t.Optional[str] = None,
        custom_auth_params: t.Optional[tool_execute_params.CustomAuthParams] = None,
        user_id: t.Optional[str] = None,
        text: t.Optional[str] = None,
        version: t.Optional[str] = None,
        modifiers: t.Optional[Modifiers] = None,
    ) -> ToolExecutionResponse:
        """
        Execute a tool with the provided parameters.

        This method calls the Composio API or a custom tool handler to execute
        the tool and returns the response. It automatically determines whether
        to use a custom tool or a Composio API tool based on the slug.

        :param slug: The slug of the tool to execute.
        :param arguments: The arguments to pass to the tool.
        :param connected_account_id: The ID of the connected account to use for the tool.
        :param custom_auth_params: The custom auth params to use for the tool.
        :param user_id: The ID of the user to execute the tool for.
        :param text: The text to pass to the tool.
        :param version: The version of the tool to execute.
        :param modifiers: The modifiers to apply to the tool.
        :return: The response from the tool.
        """
        tool = self._tool_schemas.get(slug)
        if tool is None:
            tool = t.cast(Tool, self._client.tools.retrieve(tool_slug=slug))
            self._tool_schemas[slug] = tool

        if modifiers is not None:
            processed_params = apply_modifier_by_type(
                modifiers=modifiers,
                toolkit=tool.toolkit.slug,
                tool=slug,
                type="before_execute",
                request={
                    "connected_account_id": connected_account_id,
                    "custom_auth_params": custom_auth_params,
                    "version": version,
                    "text": text,
                    "user_id": user_id,
                    "arguments": arguments,
                },
            )
            connected_account_id = processed_params["connected_account_id"]
            custom_auth_params = processed_params["custom_auth_params"]
            text = processed_params["text"]
            version = processed_params["version"]
            user_id = processed_params["user_id"]
            arguments = processed_params["arguments"]

        arguments = self._file_helper.substitute_file_uploads(
            tool=tool,
            request=arguments,
        )
        response = (
            self._execute_custom_tool(slug=slug, arguments=arguments)
            if self._custom_tools.get(slug) is not None
            else self._execute_tool(
                slug=slug,
                arguments=arguments,
                connected_account_id=connected_account_id,
                custom_auth_params=custom_auth_params,
                user_id=user_id,
                text=text,
                version=version,
            )
        )
        response = self._file_helper.substitute_file_downloads(
            tool=tool,
            response=response,
        )
        if modifiers is not None:
            response = apply_modifier_by_type(
                modifiers=modifiers,
                toolkit=tool.toolkit.slug,
                tool=slug,
                type="after_execute",
                response=response,
            )
        return response

    def proxy(
        self, **options: te.Unpack[tool_proxy_params.ToolProxyParams]
    ) -> tool_proxy_response.ToolProxyResponse:
        """Proxy a tool call to the Composio API"""
        return self._client.tools.proxy(
            endpoint=options["endpoint"],
            method=options["method"],
            body=options.get("body", self._client.not_given),
            connected_account_id=options.get(
                "connected_account_id", self._client.not_given
            ),
            parameters=options.get("parameters", self._client.not_given),
        )


__all__ = [
    "Tools",
    "ToolExecuteParams",
    "ToolExecutionResponse",
    "Modifiers",
    "Modifiers",
    "after_execute",
    "before_execute",
    "schema_modifier",
]



import json
import asyncio
import typing as t
from inspect import Signature
from typing import List, cast

import pydantic
import pydantic.error_wrappers
from agents import FunctionTool, Tool

from composio.core.provider.agentic import AgenticProviderExecuteFn
from composio.types import Tool
from composio.core.provider import AgenticProvider
from composio.utils.pydantic import parse_pydantic_error


# Recursively remove 'examples' keys from the schema properties
def _remove_examples_from_schema(schema_obj: t.Dict[str, t.Any]) -> None:
    """
    Remove 'examples', 'pattern', and 'default' keys from all properties in the
    schema, including nested ones. Also ensure that any 'items' object has a 'type' key.
    """
    # Handle properties directly
    if "properties" in schema_obj and isinstance(
        schema_obj["properties"], dict
    ):
        for _, prop_value in schema_obj["properties"].items():
            if isinstance(prop_value, dict):
                # Remove examples, pattern, and default from this property
                if "examples" in prop_value:
                    del prop_value["examples"]
                if "pattern" in prop_value:
                    del prop_value["pattern"]
                if "default" in prop_value:
                    del prop_value["default"]

                # Ensure 'items' has a 'type' key
                if "items" in prop_value and isinstance(
                    prop_value["items"], dict
                ):
                    if "type" not in prop_value["items"]:
                        # Default to string type for items if not specified
                        prop_value["items"]["type"] = "string"

                # Recursively process nested properties
                _remove_examples_from_schema(prop_value)

    # Handle array items
    if "items" in schema_obj and isinstance(schema_obj["items"], dict):
        if "examples" in schema_obj["items"]:
            del schema_obj["items"]["examples"]
        if "pattern" in schema_obj["items"]:
            del schema_obj["items"]["pattern"]
        if "default" in schema_obj["items"]:
            del schema_obj["items"]["default"]
        # Ensure items has a type
        if "type" not in schema_obj["items"]:
            schema_obj["items"]["type"] = "string"
        _remove_examples_from_schema(schema_obj["items"])

    # Handle any other nested object properties
    for _, value in schema_obj.items():
        if isinstance(value, dict):
            _remove_examples_from_schema(value)
        elif isinstance(value, list):
            for item in value:
                if isinstance(item, dict):
                    _remove_examples_from_schema(item)

class OpenAIAgentsProvider(
    AgenticProvider[FunctionTool, list[FunctionTool]],
    name="openai_agents",
):
    """
    Composio toolset for OpenAI Agents framework.
    """

    def wrap_tool(
        self,
        tool: Tool,
        execute_tool: AgenticProviderExecuteFn,
    ) -> FunctionTool:
        """Wrap a tool as a FunctionTool."""
        # Create a function that accepts explicit JSON string for parameters
        # This avoids the issue with **kwargs in schema validation
        async def execute_tool_wrapper(_ctx, payload):
            """Execute Composio action with the given arguments."""
            try:
                return json.dumps(
                    obj=(
                        await asyncio.to_thread(  # Running a thread since `execute_tool` is not async
                            execute_tool,
                            slug=tool.slug,
                            arguments=json.loads(payload) if payload else {},
                        )
                    )
                )
            except pydantic.ValidationError as e:
                return json.dumps(
                    {
                        "successful": False,
                        "error": parse_pydantic_error(e),
                        "data": None,
                    }
                )
            except Exception as e:
                return json.dumps(
                    {
                        "successful": False,
                        "error": str(e),
                        "data": None,
                    }
                )

        # Ensure the schema has additionalProperties set to false
        # this is required by OpenAI's function validation
        modified_schema = tool.input_parameters.copy()
        modified_schema["additionalProperties"] = False

        # Apply the example removal function. This is done to optimize the
        # schema as much as possible for Responses API
        _remove_examples_from_schema(modified_schema)

        # Create a custom FunctionTool with the appropriate schema
        return FunctionTool(
            name=tool.slug,
            description=tool.description,
            params_json_schema=modified_schema,
            on_invoke_tool=execute_tool_wrapper,
            # To avoid schema errors due to required flags. Composio tools
            # already process to have optimal schemas.
            strict_json_schema=False,
        )

    def wrap_tools(
        self,
        tools: t.Sequence[Tool],
        execute_tool: AgenticProviderExecuteFn,
    ) -> list[FunctionTool]:
        """Wrap a list of tools as a list of FunctionTools."""
        return [self.wrap_tool(tool, execute_tool) for tool in tools]



"""
CrewAI demo.
"""

import os

from crewai import Agent, Crew, Task
from langchain_openai import ChatOpenAI

from composio import Composio
from composio_crewai import CrewAIProvider


# Initialize tools.
openai_client = ChatOpenAI()
composio = Composio(provider=CrewAIProvider())

# Get All the tools
tools = composio.tools.get(user_id="default", toolkits=["GITHUB"])

# Define agent
crewai_agent = Agent(
    role="Github Agent",
    goal="""You take action on Github using Github APIs""",
    backstory=(
        "You are AI agent that is responsible for taking actions on Github "
        "on users behalf. You need to take action on Github using Github APIs"
    ),
    verbose=True,
    tools=tools,
    llm=openai_client,
)

# Define task
task = Task(
    description=(
        "Star a repo composiohq/composio on GitHub, if the action is successful "
        "include Action executed successfully"
    ),
    agent=crewai_agent,
    expected_output="if the star happened",
)

my_crew = Crew(agents=[crewai_agent], tasks=[task])
result = my_crew.kickoff()
print(result)



"""
Google AI Python Gemini tool spec.
"""

import typing as t

from proto.marshal.collections.maps import MapComposite
from vertexai.generative_models import (
    Content,
    FunctionDeclaration,
    GenerationResponse,
    Part,
    Tool,
)

from composio.core.provider import NonAgenticProvider
from composio.types import Modifiers, Tool, ToolExecutionResponse


def _convert_map_composite(obj):
    if isinstance(obj, MapComposite):
        return {k: _convert_map_composite(v) for k, v in obj.items()}
    if isinstance(obj, (list, tuple)):
        return [_convert_map_composite(item) for item in obj]
    return obj


class GoogleProvider(
    NonAgenticProvider[FunctionDeclaration, list[FunctionDeclaration]],
    name="google",
):
    """
    Composio toolset for Google AI Python Gemini framework.
    """

    def wrap_tool(self, tool: Tool) -> FunctionDeclaration:
        """Wraps composio tool as Google AI Python Gemini FunctionDeclaration object."""
        # Clean up properties by removing 'examples' field
        properties = t.cast(
            dict[str, dict],
            tool.input_parameters.get("properties", {}),
        )
        cleaned_properties = {
            prop_name: {k: v for k, v in prop_schema.items() if k != "examples"}
            for prop_name, prop_schema in properties.items()
        }
        return FunctionDeclaration(
            name=tool.slug,
            description=tool.description,
            parameters={
                "type": "object",
                "properties": cleaned_properties,
                "required": tool.input_parameters.get("required", []),
            },
        )

    def wrap_tools(self, tools: t.Sequence[Tool]) -> list[FunctionDeclaration]:
        return [self.wrap_tool(tool) for tool in tools]

    def execute_tool_call(
        self,
        user_id: str,
        function_call: t.Any,
        modifiers: t.Optional[Modifiers] = None,
    ) -> ToolExecutionResponse:
        """
        Execute a function call.

        :param function_call: Function call metadata from Gemini model response.
        :param entity_id: Entity ID to use for executing the function call.
        :return: Object containing output data from the function call.
        """
        return self.execute_tool(
            slug=function_call.name,
            arguments=t.cast(
                dict,
                _convert_map_composite(
                    function_call.args,
                ),
            ),
            modifiers=modifiers,
            user_id=user_id,
        )

    def handle_response(
        self,
        user_id: str,
        response: GenerationResponse,
        modifiers: t.Optional[Modifiers] = None,
    ) -> t.List[ToolExecutionResponse]:
        """
        Handle response from Google AI Python Gemini model.

        :param response: Generation response from the Gemini model.
        :param entity_id: Entity ID to use for executing the function call.
        :return: A list of output objects from the function calls.
        """
        outputs = []
        for candidate in response.candidates:
            if isinstance(candidate.content, Content) and candidate.content.parts:
                for part in candidate.content.parts:
                    if isinstance(part, Part) and part.function_call:
                        outputs.append(
                            self.execute_tool_call(
                                user_id=user_id,
                                function_call=part.function_call,
                                modifiers=modifiers,
                            )
                        )
        return outputs



from composio import Composio

composio = Composio()

# List all auth configs
auth_configs = composio.auth_configs.list()
print(auth_configs)

# Use composio managed auth
auth_config = composio.auth_configs.create(
    toolkit="github",
    options={
        "type": "use_composio_managed_auth",
    },
)
print(auth_config)

# Use custom auth
auth_config = composio.auth_configs.create(
    toolkit="notion",
    options={
        "name": "Notion Auth",
        "type": "use_custom_auth",
        "auth_scheme": "OAUTH2",
        "credentials": {
            "client_id": "1234567890",
            "client_secret": "1234567890",
            "oauth_redirect_uri": "https://backend.composio.dev/api/v1/auth-apps/add",
        },
    },
)
print(auth_config)

# When creating an auth config, you can check for required fields
required_fields = composio.toolkits.get_auth_config_creation_fields(
    toolkit="NOTION",
    auth_scheme="OAUTH2",
)
print(required_fields)

# Retrieve a specific auth config
auth_config_retrieved = composio.auth_configs.get(auth_config.id)
print(auth_config_retrieved)

# Fetch required input fields for auth config
toolkit = composio.toolkits.get(slug="notion")
required_input_fields = [
    field.name for field in toolkit.auth_config_details or [] if field.mode
]
print(required_input_fields)

# Update an auth config
auth_config_updated = composio.auth_configs.update(
    auth_config.id,
    options={
        "type": "custom",
        "credentials": {
            "client_id": "1234567890",
            "client_secret": "1234567890",
        },
        "restrict_to_following_tools": ["github"],
    },
)
print(auth_config_updated)

# Enable an auth config
composio.auth_configs.enable(auth_config.id)
print("Auth config enabled")

# Disable an auth config
composio.auth_configs.disable(auth_config.id)
print("Auth config disabled")

# Delete an auth config
composio.auth_configs.delete(auth_config.id)
print("Auth config deleted")



from .auth_configs import AuthConfigs
from .connected_accounts import ConnectedAccounts
from .toolkits import Toolkits
from .tools import Tools
from .triggers import Triggers

__all__ = [
    "AuthConfigs",
    "ConnectedAccounts",
    "Toolkits",
    "Tools",
    "Triggers",
]



from __future__ import annotations

import functools
import typing as t

import typing_extensions as te

if t.TYPE_CHECKING:
    from .tools import Tool, ToolExecutionResponse, tool_execute_params


# TODO: Maybe use `te.Unpack` in tools.execute?
class ToolExecuteParams(te.TypedDict):
    arguments: t.Dict
    connected_account_id: t.Optional[str]
    custom_auth_params: t.Optional["tool_execute_params.CustomAuthParams"]
    user_id: t.Optional[str]
    text: t.Optional[str]
    version: t.Optional[str]


ModifierInOut = t.Union["ToolExecuteParams", "ToolExecutionResponse", "Tool"]


class BeforeExecute(t.Protocol):
    """
    A modifier that is called before the tool is executed.
    """

    def __call__(
        self,
        tool: str,
        toolkit: str,
        params: ToolExecuteParams,
    ) -> ToolExecuteParams: ...


class AfterExecute(t.Protocol):
    """
    A modifier that is called after the tool is executed.
    """

    def __call__(
        self,
        tool: str,
        toolkit: str,
        response: ToolExecutionResponse,
    ) -> ToolExecutionResponse: ...


class SchemaModifier(t.Protocol):
    """
    A modifier that is called to modify the schema of the tool.
    """

    def __call__(
        self,
        tool: str,
        toolkit: str,
        schema: "Tool",
    ) -> "Tool": ...


ModifierSlug: t.TypeAlias = str
AfterExecuteModifierL: t.TypeAlias = t.Literal["after_execute"]
BeforeExecuteModifierL: t.TypeAlias = t.Literal["before_execute"]
SchemaModifierL: t.TypeAlias = t.Literal["schema"]


class Modifier:
    def __init__(
        self,
        modifier: t.Optional[AfterExecute | BeforeExecute | SchemaModifier],
        type_: AfterExecuteModifierL | BeforeExecuteModifierL | SchemaModifierL,
        tools: t.List[str],
        toolkits: t.List[str],
    ) -> None:
        self.modifier = modifier
        self.tools = tools
        self.type = type_
        self.toolkits = toolkits

    def apply(
        self,
        toolkit: str,
        tool: str,
        data: ModifierInOut,
        modifer_type: str,
    ) -> ModifierInOut:
        if self.modifier is None:
            raise ValueError("Modifier is not provided")

        # If no tools or toolkits are provided, apply the modifier to all tools
        if (
            self.type == modifer_type
            and len(self.tools) == 0
            and len(self.toolkits) == 0
        ):
            return self.modifier(tool, toolkit, data)  # type: ignore

        # If the modifier is not the same type, or the slug is not in the tools or
        # toolkits, return the data as is
        if (
            self.type != modifer_type
            or tool not in self.tools
            and toolkit not in self.toolkits
        ):
            return data

        # Apply the modifier to the data
        return self.modifier(tool, toolkit, data)  # type: ignore


@t.overload
def after_execute(
    modifier: t.Optional[AfterExecute],
) -> Modifier: ...


@t.overload
def after_execute(
    *,
    tools: t.Optional[t.List[str]] = None,
    toolkits: t.Optional[t.List[str]] = None,
) -> t.Callable[[AfterExecute], Modifier]: ...


def after_execute(
    modifier: t.Optional[AfterExecute] = None,
    *,
    tools: t.Optional[t.List[str]] = None,
    toolkits: t.Optional[t.List[str]] = None,
) -> Modifier | t.Callable[[AfterExecute], Modifier]:
    if modifier is not None:
        return Modifier(
            modifier=modifier,
            type_="after_execute",
            tools=tools or [],
            toolkits=toolkits or [],
        )

    if tools is not None or toolkits is not None:
        return t.cast(
            t.Callable[[AfterExecute], Modifier],
            functools.partial(
                after_execute,
                tools=tools or [],
                toolkits=toolkits or [],
            ),
        )

    raise ValueError("Either tools or toolkits must be provided")


@t.overload
def before_execute(modifier: t.Optional[BeforeExecute]) -> Modifier: ...


@t.overload
def before_execute(
    *,
    tools: t.Optional[t.List[str]] = None,
    toolkits: t.Optional[t.List[str]] = None,
) -> t.Callable[[BeforeExecute], Modifier]: ...


def before_execute(
    modifier: t.Optional[BeforeExecute] = None,
    *,
    tools: t.Optional[t.List[str]] = None,
    toolkits: t.Optional[t.List[str]] = None,
) -> Modifier | t.Callable[[BeforeExecute], Modifier]:
    if modifier is not None:
        return Modifier(
            modifier=modifier,
            type_="before_execute",
            tools=tools or [],
            toolkits=toolkits or [],
        )

    if tools is not None or toolkits is not None:
        return t.cast(
            t.Callable[[BeforeExecute], Modifier],
            functools.partial(
                before_execute,
                tools=tools or [],
                toolkits=toolkits or [],
            ),
        )

    raise ValueError("Either tools or toolkits must be provided")


@t.overload
def schema_modifier(modifier: t.Optional[SchemaModifier]) -> Modifier: ...


@t.overload
def schema_modifier(
    *,
    tools: t.Optional[t.List[str]] = None,
    toolkits: t.Optional[t.List[str]] = None,
) -> t.Callable[[SchemaModifier], Modifier]: ...


def schema_modifier(
    modifier: t.Optional[SchemaModifier] = None,
    *,
    tools: t.Optional[t.List[str]] = None,
    toolkits: t.Optional[t.List[str]] = None,
) -> Modifier | t.Callable[[SchemaModifier], Modifier]:
    if modifier is not None:
        return Modifier(
            modifier=modifier,
            type_="schema",
            tools=tools or [],
            toolkits=toolkits or [],
        )

    if tools is not None or toolkits is not None:
        return t.cast(
            t.Callable[[SchemaModifier], Modifier],
            functools.partial(
                schema_modifier,
                tools=tools or [],
                toolkits=toolkits or [],
            ),
        )

    raise ValueError("Either tools or toolkits must be provided")


Modifiers = t.List[Modifier]


@t.overload
def apply_modifier_by_type(
    modifiers: Modifiers,
    toolkit: str,
    tool: str,
    *,
    type: BeforeExecuteModifierL,
    request: ToolExecuteParams,
) -> ToolExecuteParams: ...


@t.overload
def apply_modifier_by_type(
    modifiers: Modifiers,
    toolkit: str,
    tool: str,
    *,
    type: AfterExecuteModifierL,
    response: "ToolExecutionResponse",
) -> "ToolExecutionResponse": ...


@t.overload
def apply_modifier_by_type(
    modifiers: Modifiers,
    toolkit: str,
    tool: str,
    *,
    type: t.Literal["schema"],
    schema: "Tool",
) -> "Tool": ...


def apply_modifier_by_type(
    modifiers: Modifiers,
    toolkit: str,
    tool: str,
    *,
    type: t.Literal["before_execute", "after_execute", "schema"],
    schema: t.Optional["Tool"] = None,
    request: t.Optional["ToolExecuteParams"] = None,
    response: t.Optional["ToolExecutionResponse"] = None,
) -> ModifierInOut:
    """Apply a modifier to a tool."""
    data = schema or request or response
    if data is None:
        raise ValueError("No data provided")

    for modifier in modifiers:
        data = modifier.apply(
            toolkit=toolkit,
            tool=tool,
            data=data,
            modifer_type=type,
        )
    return data


class ToolOptions(te.TypedDict):
    modify_schema: te.NotRequired[
        t.Dict[ModifierSlug, AfterExecute | BeforeExecute | SchemaModifier]
    ]



"""
Setup configuration for Composio OpenAI Agents plugin
"""

from pathlib import Path

from setuptools import find_packages, setup


setup(
    name="composio_openai_agents",
    version="1.0.0-rc5",
    author="Composio",
    author_email="tech@composio.dev",
    description="Use Composio to get array of strongly typed tools for OpenAI Agents",
    long_description=(Path(__file__).parent / "README.md").read_text(encoding="utf-8"),
    long_description_content_type="text/markdown",
    url="https://github.com/ComposioHQ/composio",
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: Apache Software License",
        "Operating System :: OS Independent",
    ],
    python_requires=">=3.9,<4",
    packages=find_packages(),
    install_requires=["openai-agents>=0.0.3"],
    include_package_data=True,
)



from __future__ import annotations

import abc
import typing as t

from composio.client.types import Tool
from composio.core.provider.base import BaseProvider, TTool, TToolCollection

if t.TYPE_CHECKING:
    from composio.core.models.tools import Modifiers, ToolExecutionResponse


class NoneAgenticProviderExecuteFn(t.Protocol):
    def __call__(
        self,
        slug: str,
        arguments: t.Dict,
        modifiers: t.Optional[Modifiers] = None,
        user_id: t.Optional[str] = None,
    ) -> ToolExecutionResponse:
        """
        Execute a wrapped tool by slug, passing an arbitrary input dict.
        Returns a dict with the following keys:
            - data: The data returned by the tool.
            - error: The error returned by the tool.
            - successful: Whether the tool was successful.
        """
        ...


class NonAgenticProvider(BaseProvider, t.Generic[TTool, TToolCollection]):
    """
    Base class for all non-agentic providers, such as `openai` This class is not
    meant to be used directly, but rather to be extended by concrete implementations
    This version doesn't have the execute_tool_fn for `wrap_tool` and `wrap_tools`
    """

    execute_tool: NoneAgenticProviderExecuteFn
    """The function to execute a tool. This is automatically injected by the core SDK"""

    def __init_subclass__(cls, name: str) -> None:
        cls.name = name

    def set_execute_tool_fn(self, execute_tool_fn: NoneAgenticProviderExecuteFn):
        self.execute_tool = execute_tool_fn

    @abc.abstractmethod
    def wrap_tool(
        self,
        tool: Tool,
    ) -> TTool:
        """Wrap a tool in the provider-specific format"""

    @abc.abstractmethod
    def wrap_tools(
        self,
        tools: t.Sequence[Tool],
    ) -> TToolCollection:
        """Wrap a list of tools in the provider-specific format"""



import abc
import typing as t

from composio.client.types import Tool
from composio.core.provider.base import BaseProvider, TTool, TToolCollection


class AgenticProviderExecuteFn(t.Protocol):
    def __call__(
        self,
        slug: str,
        arguments: t.Dict,
    ) -> t.Dict:
        """
        Execute a wrapped tool by slug, passing an arbitrary input dict.
        Returns a dict with the following keys:
            - data: The data returned by the tool.
            - error: The error returned by the tool.
            - successful: Whether the tool was successful.
        """
        ...


class AgenticProvider(BaseProvider, t.Generic[TTool, TToolCollection]):
    """
    Base class for all agentic providers. This class is not meant to be used
    directly but rather to be extended by concrete provider implementations.
    """

    tool_type: t.Type[TTool]
    tool_collection_type: t.Type[TToolCollection]

    def __init_subclass__(cls, name: str) -> None:
        cls.name = name

    @abc.abstractmethod
    def wrap_tool(
        self,
        tool: Tool,
        execute_tool: AgenticProviderExecuteFn,
    ) -> TTool:
        """Wrap a tool in the provider-specific format"""

    @abc.abstractmethod
    def wrap_tools(
        self,
        tools: t.Sequence[Tool],
        execute_tool: AgenticProviderExecuteFn,
    ) -> TToolCollection:
        """Wrap a list of tools in the provider-specific format"""
